# BARBARIAN_CIV_MOD
#
# by jdog5000
# Version 0.8
#


from CvPythonExtensions import *
import CvUtil
import PyHelpers		
import Popup as PyPopup
from math import sqrt
#from operator import xor

# globals
gc = CyGlobalContext()
PyPlayer = PyHelpers.PyPlayer
PyInfo = PyHelpers.PyInfo
game = CyGame()

class BarbarianCiv :

    def __init__(self, customEM, config ) :

        CvUtil.pyPrint("Initializing BarbarianCiv Mod")

        self.config = config
        
        self.EventKeyDown=6

        # Number of civs to kill at game start, to free space for new civs later
        # Python implementation only
        self.killFraction = self.config.getfloat("BarbarianCiv", "KillFraction", .25)
        
        # Radius around one city considered "close" when forming civ from barb cities
        # Reduce for smaller maps?
        self.cityRadius = self.config.getint("BarbarianCiv", "CityRadius", 10)

        # Allow barb cities to turn into minor civs
        self.allowMinor = self.config.getboolean("BarbarianCiv", "AllowMinor", False)

        # If city was founded by other than barbs, bring back a similar style civ (if possible)
        self.barbCivsByStyle = self.config.getboolean("BarbarianCiv", "BarbCivsByStyle", True)

        # Max number of turns after a militaristic barb civ starts alone will they declare war on first contact
        self.vikingWindow = self.config.getint("BarbarianCiv", "MaxVikingStyleWindow", 30)

        # popup alert when new civ is formed from barb cities
        self.LOG_BARB_SETTLE = self.config.getboolean("BarbarianCiv", "BarbSettlePopup", True)
        # enable debug log and functions
        self.LOG_BARB_DEBUG  = self.config.getboolean("BarbarianCiv", "DebugMode", False)    

        self.customEM = customEM

        #self.customEM.addEventHandler( "GameStart", self.onGameStart )
        self.customEM.addEventHandler( "EndGameTurn", self.onEndGameTurn )
        if( self.LOG_BARB_DEBUG ) : self.customEM.addEventHandler( "firstContact", self.onFirstContact )
        if( self.LOG_BARB_DEBUG ) : self.customEM.addEventHandler( "kbdEvent", self.onKbdEvent )

    def onGameStart(self, argsList) :
        'kill off some civs'
        # This is a temporary function, here to create some empty slots
        # These slots may then be filled by barb cities turning into full civs
        numToKill = 0
        if( self.killFraction <= 1 ) :
            numToKill = int( self.killFraction*(game.countCivPlayersAlive() - game.getNumHumanPlayers()) )

        if( self.LOG_BARB_DEBUG ) :
            popup = PyPopup.PyPopup()
            bodStr = 'Killing off %d civs to make space for barb civs'%(numToKill)
            popup.setBodyString( bodStr )
            popup.launch()

        loopCount = 0
        while( numToKill > 0 ) :
            loopCount += 1
            unluckyIdx = game.getSorenRandNum(gc.getMAX_CIV_PLAYERS(),'BarbarianCiv: Killing some player')
            unlucky = gc.getPlayer(unluckyIdx)
            if( unlucky.getNumUnits() > 0 and not unlucky.isHuman() ) :
                unlucky.killUnits()
                unlucky.killCities()
                numToKill -= 1
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Killed off the %s, idx: %d"%(unlucky.getCivilizationDescription(0),unluckyIdx))
            if( loopCount > 500 ) :
                break
                

    def onEndGameTurn( self, argsList ) :
        'check for worthy barb cities'
        iGameTurn = argsList[0]

        self.checkBarbCities()
        

    def onKbdEvent(self, argsList ):
        'keypress handler'
        eventType,key,mx,my,px,py = argsList

        if ( eventType == self.EventKeyDown ):
	    theKey=int(key)
	    
            # For debug or trial only
	    if( theKey == int(InputTypes.KB_B) and self.customEM.bShift and self.customEM.bCtrl ) :
                cityList = self.checkBarbCities(force = True)

    def checkBarbCities( self, force = False, forceCul = False ) :
        'Look for barb cities with neighbors or culture'
        #if( self.LOG_BARB_DEBUG ) : CyInterface().addImmediateMessage("Checking Barbarian Cities","")
        
        barbPlayer = PyPlayer( gc.getBARBARIAN_PLAYER() )

        if( barbPlayer.getNumCities() >= 1 ) :
            
            barbCityList = barbPlayer.getCityList()
            # ith entry of Tree contains list of cities close to ith entry of city list
            barbCityTree = list()
            cultureList = list()

            for i in range(0,len(barbCityList)) :

                ix = barbCityList[i].getX()
                iy = barbCityList[i].getY()
                closeCityList = list()
                closeCityList.append(barbCityList[i])
                cultureList.append(barbCityList[i].getCulture())

                # Build a list of cities close to this city
                for j in range(0,len(barbCityList)) :
                    if( not barbCityList[j] in closeCityList ) :
                        jx = barbCityList[j].getX()
                        jy = barbCityList[j].getY()
                        # I think this will catch cities close over edge of map ...
                        if( plotDistance(ix,iy,jx,jy) < self.cityRadius ) :
                            closeCityList.append(barbCityList[j])

                barbCityTree.append(closeCityList)

            # Find city with most neighbors cities
##            maxListLen = 0
##            maxListIdx = 0
##            for i,cityList in enumerate(barbCityTree) :
##                if( len(cityList) > maxListLen ) :
##                    maxListLen = len(cityList)
##                    maxListIdx = i
            
            # Find city with highest neighboring population
            maxListPop = 0
            maxListIdx = 0
            for i,cityList in enumerate(barbCityTree) :
                pop = 0
                for j in range(0,len(cityList)) :
                    pop += cityList[j].getPopulation()
                if( pop > maxListPop ) :
                    maxListPop = pop
                    maxListIdx = i

            maxCultureIdx = 0
            maxCulture = 0
            for i,culture in enumerate(cultureList) :
                if( culture >= maxCulture ) :
                    maxCulture = culture
                    maxCultureIdx = i                

            if( False and self.LOG_BARB_DEBUG ) :
                popup = PyPopup.PyPopup()
                bodStr = 'Found %d barbarian cities, max %d within %d'%(len(barbCityList),maxListLen,self.cityRadius)
                popup.setBodyString( bodStr )
                popup.launch()

            
            pop = maxListPop
            cl = int( barbCityList[i].getCulture()/2.0 + barbCityList[i].GetCy().countTotalCulture()/10.0 )
            # Odds modifier from config file
            mod = self.config.getfloat("BarbarianCiv", "Modifier", 1.00)
            # If lots of barb cities, increase odds
            numCitiesMod = int(sqrt(len(barbCityList)))
            # Change by gamespeed
            gsm = (661.0)/(game.getMaxTurns() + 1.0)
            # Odds of barb city settling down go up by era
            eraMod = gc.getActivePlayer().getCurrentEra() - game.getStartEra()/2.0

            popChance = int(mod*numCitiesMod*gsm*eraMod*1.5*(pop*pop))
            culChance = int(mod*gsm*eraMod*4*cl)

            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Barb odds: %d pop, %d cul %d"%(popChance,culChance,cl))

            
            if( popChance > (game.getSorenRandNum(999,'BarbarianCiv: Check barbs')) or force ) :
                #if( self.LOG_BARB_DEBUG ) : CyInterface().addImmediateMessage("Attempting to reincarnate civ, giving them %d cities"%(len(barbCityTree[maxListIdx])),"")
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Population barb roll succeeded, with %d cities, %d pop, forced? %d"%(len(barbCityTree[maxListIdx]),maxListPop,force))
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Current modifiers are: mod %f, ncm %d, gsm %f"%(mod,numCitiesMod,gsm))
                self.createBarbCiv( barbCityTree[maxListIdx] )
                return
            elif( cl > 0 and culChance > (game.getSorenRandNum(99,'BarbarianCiv: Check barbs')) or forceCul ) :
                # Check for cultured cities
                #if( self.LOG_BARB_DEBUG ) : CyInterface().addImmediateMessage("Attempting to reincarnate civ, giving them a cultured city","")
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Cult city barb roll succeeded, with %d adj cult, forced? %d"%(cl,forceCul))
                self.createBarbCiv( [barbCityList[maxCultureIdx]] )
                return
                
        else :
            if( force or forceCul ) :
                CyInterface().addImmediateMessage("No Barbarian cities found!","")

    def createBarbCiv( self, cityList ) :
        'Reincarnates a civ, giving them the cities in cityList'

        if( game.countCivPlayersAlive() < gc.getMAX_CIV_PLAYERS() ) :
            #if( self.LOG_BARB_DEBUG ) : CyInterface().addImmediateMessage("Creating new civ ...","")
            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Creating new civ ...")

            if( len(cityList) < 1 ) :
                return
            
            newPlayerIdx = -1
            # Put in empty slot
            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                if( not gc.getPlayer(i).isEverAlive() and (gc.getPlayer(i).getNumUnits() == 0) ) :
                    newPlayerIdx = i
                    break

            # Overwrite a civ that never managed to found a city (maybe a rebel)
            if( newPlayerIdx < 0 ) :
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - No empty slots, looking for 'fake' civs")
                for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                    if( (not gc.getPlayer(i).isAlive()) and (not gc.getPlayer(i).isFoundedFirstCity()) and (gc.getPlayer(i).getNumUnits() == 0) ) :
                        newPlayerIdx = i
                        break

            if( newPlayerIdx < 0 ) :
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - no 'fake' civs to replace, aborting barb civ creation")
                return

            cultPlayer = None
            cultPlayerID = cityList[0].GetCy().findHighestCulture()
            if( cultPlayerID >= 0 and cultPlayerID < gc.getBARBARIAN_PLAYER() ) :
                cultPlayer = gc.getPlayer( cultPlayerID )
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Civs new capital was culturally influenced by %s"%(cultPlayer.getCivilizationDescription(0)))

            # Can reincarnate as a minor civ if you want
            iMinor = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_MINOR')
            # Will cause problems when looking for capital later on, as barbs have no capital ... otherwise works
            iBarbarian = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_BARBARIAN')
            # Civs not currently in the game
            availableCivs = list()
            # Civs with similar style to cultPlayer, if they exist
            similarStyleCivs = list()
            for civType in range(0,gc.getNumCivilizationInfos()) :
                if( not civType == iBarbarian ) :
                    if( not civType == iMinor or self.allowMinor ) :
                        taken = False
                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                            if( civType == gc.getPlayer(i).getCivilizationType() ) :
                                if( gc.getPlayer(i).isAlive() or gc.getPlayer(i).isFoundedFirstCity() ) :
                                    taken = True
                                    break
                        if( not taken ) :
                            availableCivs.append(civType)
                            if( not cultPlayer == None ) :
                                if( gc.getCivilizationInfo( cultPlayer.getCivilizationType() ).getArtStyleType() == gc.getCivilizationInfo(civType).getArtStyleType() ) :
                                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Potential similar style civ: %s"%(gc.getCivilizationInfo(civType).getShortDescription(0)))
                                    similarStyleCivs.append(civType)

            if( len(availableCivs) < 1 ) :
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - ERROR: Unexpected lack of unused civ types")
                return

            if( len(similarStyleCivs) > 0 and self.barbCivsByStyle ) :
                # Create a civ using this style
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Creating a similar style civ to %s"%(cultPlayer.getCivilizationDescription(0)))
                availableCivs = similarStyleCivs
                    
            newCivIdx = availableCivs[game.getSorenRandNum(len(availableCivs),'BarbarianCiv: pick unused civ type')]

            leaderList = list()
            for leaderType in range(0,gc.getNumLeaderHeadInfos()) :
                if( gc.getCivilizationInfo(newCivIdx).isLeaders(leaderType) ) :
                    leaderList.append(leaderType)

            if( len(leaderList) < 1 ) :
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - ERROR: Unexpected lack of possible leaders")
                return
            
            newLeaderIdx = leaderList[game.getSorenRandNum(len(leaderList),'BarbarianCiv: pick leader')]
                
            #if( self.LOG_BARB_DEBUG ) : CyInterface().addImmediateMessage("Attempting to create the %s at idx: %d, with leader %s"%( gc.getCivilizationInfo(newCivIdx).getShortDescription(0),newPlayerIdx,gc.getLeaderHeadInfo(newLeaderIdx).getLeaderHead()),"")
       
            ix = cityList[0].getX()
            iy = cityList[0].getY()
            cityPlot = cityList[0].plot()
            # SDK versio, -1 for put on new team
            success = game.createNewPlayer( newPlayerIdx, newCivIdx, newLeaderIdx, -1, True, ix, iy )

            if( not success ) :
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - FAILED to reincarnate civ type %d in slot %d"%(newCivIdx,newPlayerIdx))
                return

            if( gc.getPlayer(newPlayerIdx).getNumCities() > 0 ) :

                newPlayer = gc.getPlayer(newPlayerIdx)
                pyNewPlayer = PyPlayer(newPlayerIdx)
                civName = newPlayer.getCivilizationDescription(0)
                leadName = newPlayer.getName()

                if( self.LOG_BARB_SETTLE or self.LOG_BARB_DEBUG ) :
                    #CyInterface().addImmediateMessage("%s has organized the Barbarians and formed the %s!"%(leadName,civName),'')
                    popup = PyPopup.PyPopup()
                    bodStr = "Word spreads from distant lands:\n\n%s has organized the Barbarians and formed the %s!"%(leadName,civName)
                    popup.setBodyString( bodStr )
                    popup.launch()
                    
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - %s has organized the Barbarians and formed the %s!  Turn: %d, Year: %d"%(leadName,civName,game.getElapsedGameTurns(),game.getTurnYear(game.getElapsedGameTurns())))
                                                      
                if( len(cityList) > 1 ) :
                    #if( self.LOG_BARB_DEBUG ) : CyInterface().addImmediateMessage("Giving the %s %d additional cities"%(civName,len(cityList)-1),"")
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Giving the %s %d additional cities"%(civName,len(cityList)-1))
                    for i,giftCity in enumerate(cityList[1:]) :
                        giftCity.changePopulation(1)
                        newPlayer.acquireCity( giftCity.GetCy(), False, False )

                cityList = pyNewPlayer.getCityList()
                capital = pyNewPlayer.getCapitalCity()
                # Find nearby civs
                newPlayerArea = capital.GetCy().area()
                closePlayers = list()
                for idx in range(0,gc.getMAX_CIV_PLAYERS()) :
                    if( not idx == newPlayerIdx ) :
                        if( newPlayerArea.getCitiesPerPlayer(idx) > 0 ) :
                            closePlayers.append(idx)

                # Pick a style for barbarian organization
                militaryOdds = int( 100*self.config.getfloat("BarbarianCiv", "MilitaryBuildupBaseOdds", 0.50) )
                iAgg = CvUtil.findInfoTypeNum(gc.getTraitInfo,gc.getNumTraitInfos(),'TRAIT_AGGRESSIVE')
                iExp = CvUtil.findInfoTypeNum(gc.getTraitInfo,gc.getNumTraitInfos(),'TRAIT_EXPANSIVE')
                if( gc.getLeaderHeadInfo( newPlayer.getLeaderType() ).hasTrait( iAgg ) ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - %s is aggressive"%(leadName))
                    militaryOdds += 25
                elif( gc.getLeaderHeadInfo( newPlayer.getLeaderType() ).hasTrait( iExp ) ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - %s is expansionist"%(leadName))
                    militaryOdds += 15
                if( len(closePlayers) == 0 ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - New civ is all alone")
                    militaryOdds -= 15

                barbSettleType = 0
                iRand = game.getSorenRandNum(99,'BarbarianCiv: Check barbs')
                if( militaryOdds > iRand ) :
                    # Military consolidation
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Military buildup, roll: %d > %d"%(militaryOdds,iRand))
                    barbSettleType = 1
                else :
                    # Builder consolidation
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Builder buildup, roll: %d <= %d"%(militaryOdds,iRand))
                    barbSettleType = 2

                # Giving the new civ free stuff!
                # Base stuff
                newPlayer.changeGold( 150 + game.getSorenRandNum(150,'BarbarianCiv: give money') )
                
                barbTechFrac = self.config.getfloat("BarbarianCiv", "BarbTechFrac", 0.60)
                for techID in range(0,gc.getNumTechInfos()) :
                    numTeams = game.countCivTeamsAlive()
                    numTeamsWhoKnow = game.countKnownTechNumTeams(techID)
                    fracKnow = numTeamsWhoKnow/(1.0*numTeams)
                    if( fracKnow >= barbTechFrac and newPlayer.canEverResearch(techID) ) :
                        if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Giving %s"%(PyInfo.TechnologyInfo(techID).getDescription()))
                        gc.getTeam(newPlayer.getTeam()).setHasTech(techID,True,newPlayerIdx,False,False)

                # One great person: merhcant, artist, prophet, scientific, or engineer
                # For Warlords: Add great general, make most likely for military buildup
                iProphet = CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_PROPHET')
                iArtist = CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_ARTIST')     # Will seemingly always be used for culture
                iMerchant = CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_MERCHANT')
                iScientist = CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_SCIENTIST')
                iEngineer = CvUtil.findInfoTypeNum(gc.getUnitInfo,gc.getNumUnitInfos(),'UNIT_ENGINEER')

                if( barbSettleType == 1 ) : gpList = [iProphet,iProphet,iScientist,iScientist,iEngineer]
                else : gpList = [iProphet,iProphet,iArtist,iMerchant,iScientist,iScientist,iEngineer,iEngineer]
                gpType = gpList[game.getSorenRandNum(len(gpList),'BarbarianCiv: pick leader')]
                newPlayer.initUnit(gpType,ix,iy,UnitAITypes.NO_UNITAI)
                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - GP type %s"%(PyInfo.UnitInfo(gpType).getDescription()))
                
                # Special stuff depending on type
                if( barbSettleType == 1 ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Military buildup!")

                    # Always give them horseback riding
                    iHorseback = CvUtil.findInfoTypeNum(gc.getTechInfo,gc.getNumTechInfos(),'TECH_HORSEBACK_RIDING')
                    if( not gc.getTeam(newPlayer.getTeam()).isHasTech(iHorseback) ) :
                        gc.getTeam(newPlayer.getTeam()).setHasTech(iHorseback,True,newPlayerIdx,False,False)
                        if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Giving military %s"%(PyInfo.TechnologyInfo(iHorseback).getDescription()))
                
                elif( barbSettleType == 2 ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Builder buildup!")

                    # Extra techs
                    for techID in range(0,gc.getNumTechInfos()) :
                        numTeams = game.countCivTeamsAlive()
                        numTeamsWhoKnow = game.countKnownTechNumTeams(techID)
                        fracKnow = numTeamsWhoKnow/(1.0*numTeams)
                        techMod = self.config.getfloat("BarbarianCiv", "BuilderTechModifier", 1.00)
                        if( not gc.getTeam(newPlayer.getTeam()).isHasTech(techID) and newPlayer.canEverResearch(techID) and techMod*100*fracKnow > game.getSorenRandNum(80,'BarbarianCiv: extra tech') ) :
                            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Giving builder %s"%(PyInfo.TechnologyInfo(techID).getDescription()))
                            gc.getTeam(newPlayer.getTeam()).setHasTech(techID,True,newPlayerIdx,False,False)

                # Basic units, find the (potentially unique) unit type for the class for this civ
                workerClass = CvUtil.findInfoTypeNum(gc.getUnitClassInfo,gc.getNumUnitClassInfos(),'UNITCLASS_WORKER')
                iWorker = gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(workerClass)
                settlerClass = CvUtil.findInfoTypeNum(gc.getUnitClassInfo,gc.getNumUnitClassInfos(),'UNITCLASS_SETTLER')
                iSettler = gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(settlerClass)
                scoutClass = CvUtil.findInfoTypeNum(gc.getUnitClassInfo,gc.getNumUnitClassInfos(),'UNITCLASS_SCOUT')
                iScout = gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(scoutClass)

                # Military units, have to vary depending on available techs
                iBestDefender = UnitTypes.NO_UNIT
                iCounter = UnitTypes.NO_UNIT
                iMobile = UnitTypes.NO_UNIT
                iAttack = UnitTypes.NO_UNIT
                iAssaultShip = UnitTypes.NO_UNIT
                for unitID in range(0,gc.getNumUnitInfos()) :
                    if( newPlayer.canTrain(unitID,False,False) and gc.getUnitInfo(unitID).getDomainType() == DomainTypes.DOMAIN_LAND ): # canTrain only does technology check
                        unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                        # Defender (Archer,Longbow)
                        if( gc.getUnitInfo(unitID).getDefaultUnitAIType() == UnitAITypes.UNITAI_CITY_DEFENSE ):
                            if( (iBestDefender == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() >= gc.getUnitInfo(iBestDefender).getCombat() ) :
                                #if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Civ def units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                if( unitID == gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                    iBestDefender = unitID
                                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Best defender set to %s"%(PyInfo.UnitInfo(iBestDefender).getDescription()))
                        # Counter (Axemen,Phalanx)
                        if( gc.getUnitInfo(unitID).getDefaultUnitAIType() == UnitAITypes.UNITAI_COUNTER ):
                            if( (iCounter == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() >= gc.getUnitInfo(iCounter).getCombat() ) :
                                #if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Civ counter units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                if( unitID == gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                    iCounter = unitID
                                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Best counter unit set to %s"%(PyInfo.UnitInfo(iCounter).getDescription()))
                        # Assault units
                        if( gc.getUnitInfo(unitID).getUnitAIType( UnitAITypes.UNITAI_ATTACK ) ):
                            if( (iAttack == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() > gc.getUnitInfo(iAttack).getCombat() ) :
                                #if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Civ attack units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                if( unitID == gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                    iAttack = unitID
                                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Best attack set to %s"%(PyInfo.UnitInfo(iAttack).getDescription()))
                        # Fast Attack (Horse archer)
                        if( gc.getUnitInfo(unitID).getUnitAIType( UnitAITypes.UNITAI_ATTACK ) and gc.getUnitInfo(unitID).getMoves() > 1 ):
                            if( (iMobile == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() > gc.getUnitInfo(iMobile).getCombat() ) :
                                #if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Civ mob units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                if( unitID == gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                    iMobile = unitID
                                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Best mobile set to %s"%(PyInfo.UnitInfo(iMobile).getDescription()))

                    elif( newPlayer.canTrain(unitID,False,False) and gc.getUnitInfo(unitID).getDomainType() == DomainTypes.DOMAIN_SEA ):
                        unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                        # Military unit transport ship
                        if( gc.getUnitInfo(unitID).getUnitAIType( UnitAITypes.UNITAI_ASSAULT_SEA ) ):
                            if( unitID == gc.getCivilizationInfo( newPlayer.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                iAssaultShip = unitID
                                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Best assault ship set to %s"%(PyInfo.UnitInfo(iAssaultShip).getDescription()))

                        
                # For Warlords: change to give unique buidlings
                iLibrary = CvUtil.findInfoTypeNum(gc.getBuildingInfo,gc.getNumBuildingInfos(),'BUILDING_LIBRARY')
                iGranary = CvUtil.findInfoTypeNum(gc.getBuildingInfo,gc.getNumBuildingInfos(),'BUILDING_GRANARY')
                iBarracks = CvUtil.findInfoTypeNum(gc.getBuildingInfo,gc.getNumBuildingInfos(),'BUILDING_BARRACKS')
                iMarket = CvUtil.findInfoTypeNum(gc.getBuildingInfo,gc.getNumBuildingInfos(),'BUILDING_MARKET')
                iWalls = CvUtil.findInfoTypeNum(gc.getBuildingInfo,gc.getNumBuildingInfos(),'BUILDING_WALLS')
                iLighthouse = CvUtil.findInfoTypeNum(gc.getBuildingInfo,gc.getNumBuildingInfos(),'BUILDING_LIGHTHOUSE')
                iForge = CvUtil.findInfoTypeNum(gc.getBuildingInfo,gc.getNumBuildingInfos(),'BUILDING_FORGE')

                # Put stuff in cities
                for i,city in enumerate(cityList) :
                    cityX = city.getX()
                    cityY = city.getY()
                    if( not iBestDefender == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iBestDefender,cityX,cityY,3)
                    if( not iWorker == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iWorker,cityX,cityY,1)
                    if( city.getPopulation() < 3 ) :
                        city.setPopulation(3)
                    if( city.getPopulation() < 6 ) :
                        city.changePopulation( game.getSorenRandNum(7-city.getPopulation(),'BarbarianCiv: change city population') )
                        if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - New population of %s is %d"%(city.getName(),city.getPopulation()))
                    if( barbSettleType == 1 ) : # Military
                        extraUnits = int( (gc.getActivePlayer().getCurrentEra()+1)/2.0 )
                        if( len(closePlayers) == 0 ) : # Viking style, no horsies (but still allows elephants ... )
                            if( not iCounter == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iCounter,cityX,cityY,2+game.getSorenRandNum(5+extraUnits,'BarbarianCiv: give military infantry'))
                            if( not iAttack == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iAttack,cityX,cityY,2+game.getSorenRandNum(5+extraUnits,'BarbarianCiv: give military mobile'))
                        else :
                            if( not iCounter == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iCounter,cityX,cityY,1+game.getSorenRandNum(4+extraUnits,'BarbarianCiv: give military infantry'))
                            if( not iMobile == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iMobile,cityX,cityY,1+game.getSorenRandNum(4+extraUnits,'BarbarianCiv: give military mobile'))
                            if( not iAttack == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iAttack,cityX,cityY,1+game.getSorenRandNum(4+extraUnits,'BarbarianCiv: give military mobile'))
                        if( city.canConstruct( iBarracks ) ) :
                            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iBarracks).getDescription(),city.getName()))
                            city.setHasRealBuildingIdx(iBarracks,True)
                    elif( barbSettleType == 2 ) : # Builder
                        if( city.canConstruct( iLibrary ) ) :
                            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iLibrary).getDescription(),city.getName()))
                            city.setHasRealBuildingIdx(iLibrary,True)
                        if( city.canConstruct( iLighthouse ) ) :
                            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iLighthouse).getDescription(),city.getName()))
                            city.setHasRealBuildingIdx(iLighthouse,True)
                        if( city.canConstruct( iGranary ) ) :
                            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iGranary).getDescription(),city.getName()))
                            city.setHasRealBuildingIdx(iGranary,True)
                    iTotCul = city.GetCy().countTotalCulture()
                    if( iTotCul > 0 ) :
                        city.setCulture( int(iTotCul/2.0) )
                        if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Culture of %s set to %d"%(city.getName(),city.getCulture()))
                        
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - %d units placed in %s"%(city.plot().getNumUnits(),city.getName()))
                        
                # More stuff in capital
                if( not iScout == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iScout,capital.getX(),capital.getY(),1)
                if( capital.canConstruct( iWalls ) ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iWalls).getDescription(),capital.getName()))
                    capital.setHasRealBuildingIdx(iWalls,True)
                if( capital.canConstruct( iForge ) ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iForge).getDescription(),capital.getName()))
                    capital.setHasRealBuildingIdx(iForge,True)
                if( city.canConstruct( iGranary ) ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iGranary).getDescription(),city.getName()))
                    city.setHasRealBuildingIdx(iGranary,True)
                if( len(cityList) < 3 ) :
                    if( not iSettler == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iSettler,capital.getX(),capital.getY(),1)
                    if( not iBestDefender == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iBestDefender,capital.getX(),capital.getY(),1)
                if( barbSettleType == 2 ) : # Builder
                    if( not iWorker == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iWorker,capital.getX(),capital.getY(),1)
                    if( not iSettler == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iSettler,capital.getX(),capital.getY(),1)
                    if( not iBestDefender == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iBestDefender,capital.getX(),capital.getY(),1)
                    if( not iCounter == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iCounter,capital.getX(),capital.getY(),2)
                    if( capital.canConstruct( iMarket ) ) :
                        if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Constructing %s in %s"%(PyInfo.BuildingInfo(iMarket).getDescription(),capital.getName()))
                        capital.setHasRealBuildingIdx(iMarket,True)
                elif( barbSettleType == 1 and len(cityList) == 1 ) : # Military extra stuff if small
                    if( not iCounter == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iCounter,cityX,cityY,1)
                    if( not iMobile == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iMobile,cityX,cityY,2)
                    if( not iAttack == UnitTypes.NO_UNIT ) : pyNewPlayer.initUnit(iAttack,cityX,cityY,1)


                # Give units some experience
                defenseUnits = pyNewPlayer.getUnitsOfType( iBestDefender )
                for unit in defenseUnits :
                    if( 60 > game.getSorenRandNum(100,'BarbarianCiv: unit experience') ) :
                        unit.setExperience( game.getSorenRandNum(7,'BarbarianCiv: unit experience'), 10 )
                if( barbSettleType == 1 ) : # Military
                    mobileUnits = pyNewPlayer.getUnitsOfType( iMobile )
                    for unit in mobileUnits :
                        unit.setExperience( game.getSorenRandNum(7,'BarbarianCiv: unit experience'), 10 )
                    counterUnits = pyNewPlayer.getUnitsOfType( iCounter )
                    for unit in counterUnits :
                        unit.setExperience( game.getSorenRandNum(7,'BarbarianCiv: unit experience'), 10 )
                    attackUnits = pyNewPlayer.getUnitsOfType( iAttack )
                    for unit in attackUnits :
                        unit.setExperience( game.getSorenRandNum(7,'BarbarianCiv: unit experience'), 10 )

                # Diplomacy
                if( len(closePlayers) > 0 and barbSettleType == 1 ) :
                    # Military buildup will declare war on some nearby civ
                    vr = game.getSorenRandNum(len(closePlayers),'BarbarianCiv: pick who to declare war on')
                    victimIdx = closePlayers[vr]
                    pyNewPlayer.getTeam().declareWar(victimIdx,True)
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - New civ declared war on %s"%(PyPlayer(victimIdx).getCivilizationName()))
                    
                elif( len(closePlayers) == 0 and barbSettleType == 1 ) :
                    # Give them some boats to go raiding!
                    for i,city in enumerate(cityList) :
                        if( city.GetCy().isCoastal() ) :
                            if( not iAssaultShip == UnitTypes.NO_UNIT ) :
                                if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Giving new civ some boats for raiding")
                                pyNewPlayer.initUnit(iAssaultShip,city.getX(),city.getY(),2)
                            break
                        
##            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
##                if( not gc.getPlayer(i).isAlive() and gc.getPlayer(i).isEverAlive() ) :
##                    newPlayerIdx = i
##                    break

                        # Python only
            #gc.getPlayer(newPlayerIdx).acquireCity( cityList[0].GetCy(),False,False)
            #cityPlot.setOwner(newPlayerIdx)
            #success = True
            
        else :
            #if( self.LOG_BARB_DEBUG ) : CyInterface().addImmediateMessage("Game is full ...","")
            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Game is full ...")


    def onFirstContact( self, argsList ) :
        'Contact'
	iTeamX,iHasMetTeamY = argsList

        if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - %d has just met %d"%(iTeamX,iHasMetTeamY))

        teamX = gc.getTeam(iTeamX)
        if( teamX.getNumMembers() > 1 ) :
            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Team %d has too many members: %d"%(iTeamX,teamX.getNumMembers()))          
            return
        
        playerX = None
        playerY = None
        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
            player = gc.getPlayer(i)
            if( player.isAlive() ):
                if( player.getTeam() == iTeamX ) :
                    playerX = player
                elif( player.getTeam() == iHasMetTeamY ) :
                    playerY = player

        if( playerX == None or playerY == None ) :
            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Hmmm ... didn't find any players")
            return

        if( playerX.getCapitalCity().getOriginalOwner() == gc.getBARBARIAN_PLAYER() ) :
            for idx in range(0,gc.getMAX_CIV_PLAYERS()) :
                if( playerX.canContact( idx ) and not idx == playerY.getID() ) :
                    if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - Barb civ %d can also contact %d, no Viking action"%(iTeamX,idx))
                    return

            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - The %s has just met its first civ, %d"%(playerX.getCivilizationDescription(0),iHasMetTeamY))
	
            if( game.getGameTurn() < (playerX.getCapitalCity().getGameTurnAcquired() + self.vikingWindow) ) :
                if( playerX.getNumMilitaryUnits()/playerX.getNumCities() > 5 ) :
                    
                    unitList = PyPlayer(playerX.getID()).getUnitList()
                    for unit in unitList :
                        if( unit.isCargo() ) :
                            if( self.LOG_BARB_DEBUG ) : CvUtil.pyPrint("  BC - The %s is going Viking on the %s!!!"%(playerX.getCivilizationDescription(0),playerY.getCivilizationDescription(0)))
                            teamX.declareWar(iHasMetTeamY,True)
                            return
                
