# ChangePlayer Mod
#
# by jdog5000
#


from CvPythonExtensions import *
import CvUtil
import PyHelpers		
import Popup as PyPopup
from operator import xor

# globals
gc = CyGlobalContext()
PyPlayer = PyHelpers.PyPlayer
PyInfo = PyHelpers.PyInfo
game = CyGame()

class ChangePlayer :

    def __init__(self, customEM, config ) :

        CvUtil.pyPrint("Initializing ChangePlayer Mod")

        self.LOG_DEBUG = config.getboolean("ChangePlayer", "DebugMode", False)

        self.EventKeyDown=6
        self.changePlayerPopup = 7002
        self.changeHumanPopup = 7003

        self.config = config
        self.customEM = customEM

        self.customEM.addEventHandler( "kbdEvent", self.onKbdEvent )

        self.customEM.setPopupHandler( self.changePlayerPopup, ["changePlayerPopup",self.changePlayerHandler,self.blankHandler] )
        self.customEM.setPopupHandler( self.changeHumanPopup, ["changeHumanPopup",self.changeHumanHandler,self.blankHandler] )

    def blankHandler( self, playerID, netUserData, popupReturn ) :
        # Dummy handler to take the second event for popup
        return


    def onKbdEvent(self, argsList ):
        'keypress handler'
        eventType,key,mx,my,px,py = argsList

        if ( eventType == self.EventKeyDown ):
	    theKey=int(key)
	    
	    if( theKey == int(InputTypes.KB_P) and self.customEM.bShift and self.customEM.bCtrl ) :
                self.changePlayer( )

            if( theKey == int(InputTypes.KB_H) and self.customEM.bShift and self.customEM.bCtrl ) :
                self.changeHuman( )

    def changePlayer( self ) :
        'Chooser window for switching a players civ'

        popup = PyPopup.PyPopup(self.changePlayerPopup,contextType = EventContextTypes.EVENTCONTEXT_ALL)
        popup.setHeaderString( 'Change a civ?' )
        popup.setBodyString( 'Which civ to change, to which civ and leader and what team?' )
        popup.addSeparator()
        #popup.createPythonEditBox( '10', 'Number of turns to turn over to AI', 0)
        
        popup.createPythonPullDown( 'Switch this civ ...', 1 )
        for i in range(0,gc.getMAX_PLAYERS()) :
            player = PyPlayer(i)
            if( not player.isNone() ) :
                if( player.isAlive() ) :
                    popup.addPullDownString( "%s of the %s"%(player.getName(),player.getCivilizationName()), i, 1 )

        activePlayer = gc.getActivePlayer()
        popup.popup.setSelectedPulldownID( activePlayer.getID(), 1 )

        popup.createPythonPullDown( ' ... to this civ', 2 )
        for i in range(0,gc.getNumCivilizationInfos()) :
            newCivInfo = gc.getCivilizationInfo(i)
            popup.addPullDownString( "%s Empire"%(newCivInfo.getAdjective(0)), i, 2 )

        popup.popup.setSelectedPulldownID( activePlayer.getCivilizationType(), 2 )

        popup.createPythonPullDown( ' ... with this leader', 3 )
        for i in range(0,gc.getNumLeaderHeadInfos()) :
            newLeaderInfo = gc.getLeaderHeadInfo(i) 
            leaderName = newLeaderInfo.getLeaderHead()  # this is the full path of the head animation
            if( leaderName.count('/') > 0 ) :
                leaderName = leaderName.split('/')[-2]  # hack to get to just name
                leaderName = leaderName.replace('_',' ')
                leaderName = leaderName.title()
            #if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : Leadername %s"%(leaderName) )

            popup.addPullDownString( "%s"%(leaderName), i, 3 )

        popup.popup.setSelectedPulldownID( activePlayer.getLeaderType(), 3 )

        popup.createPythonPullDown( ' ... on this team', 4 )
        popup.addPullDownString( "Keep current team", -1, 4 )  # Team idx of -1 maintains current team setting
        for i in range(0,gc.getMAX_PLAYERS()) :
            player = PyPlayer(i)
            if( not player.isNone() ) :
                if( player.isAlive() ) :
                    popup.addPullDownString( "Team with the %s"%(player.getCivilizationName()), i, 4 )

        popup.popup.setSelectedPulldownID( -1, 4 )
        
        popup.addSeparator()

        popup.addButton('Cancel')

        popup.launch()

    def changePlayerHandler( self, playerID, netUserData, popupReturn ) :
        'Handles ChangePlayer popup'

        if( popupReturn.getButtonClicked() == 0 ):  # if you pressed cancel
            return

        playerIdx = popupReturn.getSelectedPullDownValue( 1 )
        newCivType = popupReturn.getSelectedPullDownValue( 2 )
        newLeaderType = popupReturn.getSelectedPullDownValue( 3 )
        teamIdx = popupReturn.getSelectedPullDownValue( 4 )
     
        if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : You have selected player %d, the %s, on team %d"%(playerIdx, gc.getPlayer(playerIdx).getCivilizationDescription(0), gc.getPlayer(playerIdx).getTeam()) )
        if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : New team idx is %d"%(teamIdx) )

        player = gc.getPlayer(playerIdx)      
        game.changePlayer( playerIdx, newCivType, newLeaderType, teamIdx, player.isHuman(), True )

        CyInterface().addImmediateMessage("Player %d has been changed"%(playerIdx),"")
        if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : Player change completed" )

    def changeHuman( self ) :
        'Chooser window for switching human player'
        popup = PyPopup.PyPopup(self.changeHumanPopup,contextType = EventContextTypes.EVENTCONTEXT_ALL)
        popup.setHeaderString( 'Pick a new civ' )
        popup.setBodyString( 'Which civ would you like to control?' )
        popup.addSeparator()
        #popup.createPythonEditBox( '10', 'Number of turns to turn over to AI', 0)
        
        popup.createPythonPullDown( 'Take control of this civ ...', 1 )
        for i in range(0,gc.getMAX_PLAYERS()) :
            player = PyPlayer(i)
            if( not player.isNone() ) :
                if( player.isAlive() ) :
                    popup.addPullDownString( "%s of the %s"%(player.getName(),player.getCivilizationName()), i, 1 )

        activePlayerIdx = gc.getActivePlayer().getID()
        popup.popup.setSelectedPulldownID( activePlayerIdx, 1 )
        
        popup.addSeparator()
 
        popup.addButton('Cancel')
        
        popup.launch()

    def changeHumanHandler( self, playerID, netUserData, popupReturn ) :
        'Handles changeHuman popup'
        
        if( popupReturn.getButtonClicked() == 0 ):  # if you pressed cancel
            return
        
        newHumanIdx = popupReturn.getSelectedPullDownValue( 1 )
        newPlayer = gc.getPlayer(newHumanIdx)
        oldHumanIdx = playerID
        oldPlayer = gc.getPlayer(oldHumanIdx)

        if( newHumanIdx == oldHumanIdx ) :
            if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : You have selected the same civ, no change")
            CyInterface().addImmediateMessage("You retain control of the %s"%(oldPlayer.getCivilizationDescription(0)),"")
            return
        
        
        if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : You have selected player %d, the %s"%(newHumanIdx, newPlayer.getCivilizationDescription(0)) )
        
        game.setActivePlayer( newHumanIdx, False )
        game.changePlayer( newHumanIdx, newPlayer.getCivilizationType(), newPlayer.getLeaderType(), -1, True, False )
        game.changePlayer( oldHumanIdx, oldPlayer.getCivilizationType(), oldPlayer.getLeaderType(), -1, False, False )

        if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : Number of human players is now %d"%(game.getNumHumanPlayers()) )

        CyInterface().addImmediateMessage("You now control the %s"%(newPlayer.getCivilizationDescription(0)),"")
