# Rebellion Mod
# 
# Originally by Jon 'Trip' Shafer - 05 Nov 2005
# Civilization 4 (c) 2005 Firaxis Games
#
# Adapted by jdog5000
# Version 0.5
#
# Have fun!

from CvPythonExtensions import *
import sys
import Popup as PyPopup
from PyHelpers import PyPlayer
import pickle
import CvEventManager
from PyHelpers import *
import CvUtil
from math import sqrt
from CvCameraControls import CvCameraControls

# globals
gc = CyGlobalContext()
localText = CyTranslator()
game = CyGame()

DefaultUnitAI = UnitAITypes.NO_UNITAI

class Rebellion :

	def __init__(self, customEM, config):

                CvUtil.pyPrint("Initializing Rebellion Mod")
		
                self.LOG_DEBUG = config.getboolean("Rebellion", "DebugMode", False)
        		
		####### System Variables #######
		
		# The City Script array IDs when loading/saving data
		self.iCitySpawnScriptID = 0
		
		####### Game Variables #######
		
		self.iBarbariansID = gc.getBARBARIAN_PLAYER()
		self.rebelOfferPopup = 7005
		self.iBuyoffCost = 0
		self.pCity = None
		self.iTotalRebelValue = 0
		
		# Increase this value to make a Revolution more likely
		self.iRevolutionMod = 0#0
		# Chance of Rebel unit spawning near a city while the civ is in a Revolution
		self.iRebelSpawnRoll = config.getint("Rebellion", "Roll", 10000) #10000
		# Range from cities rebel units can spawn
		self.iUnitSpawnRange = 2#2
		# Number of turns between when rebel units can spawn
		self.iCitySpawnUnitCounterMax = config.getint("Rebellion", "Turns", 8)#6
		# Threshold below which there is no chance of rebellion
		self.iThreshold = config.getint("Rebellion", "Threshold", 400)
		# If true, cities with no unhappy citizens can also rebel
		self.allowHappyRebellions = config.getboolean("Rebellion", "AllowHappyRebellions", False)
		# If True, spawn rebels as Barbarians, otherwise try to spawn a new rebel player
		self.rebelsAsBarbs = config.getboolean("Rebellion", "RebelsAsBarbs", False)
		
		############## CITY VS STATE RELIGION ##############
		
		# Modifies how much a city having a different religion from the state affects chance of rebels
		self.iDifferentReligionFactor = 100 # Default is 150
		# Increase different religion factor when running a theocracy
		self.iTheocracyMultiplier = 4
		# Increase different religion factor when running Organized religion
		self.iOrganizedReligionMultiplier = 2
		
		############## CIVICS ##############
				
		# Modifies how much a player's government civic option affects chance of rebels
		self.iDespotismFactor = 50 # Default is 50
		self.iHereditaryRuleFactor = 50 # Default is 50
		self.iPoliceStateFactor = 100 # Default is 100
		self.iUniversalSuffrageFactor = -50
		
		# Modifies how much a player's legal civic option affects chance of rebels
		self.iBarbarismFactor = 50 # Default is 50
		self.iVassalageFactor = 75 # Default is 75
		
		# Modifies how much a player's labor civic option affects chance of rebels
		self.iTribalismFactor = 50 # Default is 50
		self.iSlaveryFactor = 100 # Default is 100
		self.iSerfdomFactor = 100 # Default is 100
		self.iCasteSystemFactor = 85 # Default is 85
		
		# Modifies how much a player's religion civic option affects chance of rebels
		self.iPaganismFactor = 25 # Default is 50
		# Most of effect for these civics is factored into religion factor
		self.iOrganizedReligionFactor = 25#100 # Default is 100
		self.iTheocracyFactor = 50#100 # Default is 100
		
		############## LAND AREA ##############
		
		# Modifies how much a player's total amount of land affects chance of rebels
		self.fLandModifier = config.getfloat("Rebellion", "LandFactor", 1.00)
		
		############## HEALTH AND HAPPINESS ##############
		
		# Modifies how much a player's city unhealthiness affects chance of rebels
		self.iHealthFactor = 200 # Default is 200
		self.iHealthyDivisor = 8
		self.iHappinessFactor = 400 # Default is 300
		self.iHappyDivisor = 6
		
		############## CITY NATIONALITY ##############
		
		self.iNationalityFactor = 1000 # Default is 1000
		
		############## CIV SIZE ##############
		
		# Modifies how much a player's total land area affects chance of rebels
		self.iCivSizeFactor = 500 # Default is 1000
		# When a tiny civ has a distant colony, this caps the size of the modifier
		self.maxCityDistModifier = 3.0

		############## CITY AREA ###############

		# Modifies how much city being in a different area than capital affects chance of rebels
		self.iCityAreaFactor = 300
		
		############## ? ##############
		
		# Modifies how much a  affects chance of rebels
		self.fModifier = config.getfloat("Rebellion", "Modifier", 1.00)
		self.fSizeModifier = config.getfloat("Rebellion", "RebellionSizeModifier", 1.00)
		
		self.szEventText = ""
		self.szResultText = ""

		self.customEM = customEM

		# Register Rebellion events:
		self.customEM.addEventHandler( "GameStart", self.onGameStart )
		self.customEM.addEventHandler( "OnLoad", self.onLoadGame )
		self.customEM.addEventHandler( "EndGameTurn", self.onEndGameTurn )
		self.customEM.addEventHandler( "cityBuilt", self.onCityBuilt )
		self.customEM.addEventHandler( "cityAcquired", self.onCityAcquired )

		self.customEM.setPopupHandler( self.rebelOfferPopup, ["rebelOfferPopup",self.rebelOfferHandler,self.blankHandler] )

		#self.initValues()

	def blankHandler( self, playerID, netUserData, popupReturn ) :
            # Dummy handler to take the second event for popup
            return
		
	def initValues(self):
		
		self.iWarriorID = CvUtil.findInfoTypeNum(gc.getUnitInfo, gc.getNumUnitInfos(), 'UNIT_WARRIOR')
		
	def turnChecker(self, iTurn):
		
		self.szGameDate = CyGameTextMgr().getTimeStr(iTurn, false)
		
		# Alter City Unit Spawning Counters
		self.alterCitySpawnUnitCounters()
		
		# Determine where rebels show up and whatnot
		self.checkForRebellion(iTurn)

		# Remove floating rebels (ie rebels at peace, but have no cities)
		self.removeFloatingRebellions()
		
###################### Events ####################################

        def onGameStart(self, argsList):
		'Called at the start of the game'
		
		# Set up mod-specific values
		self.initValues()
	
	def onLoadGame(self, argsList):
		'Called when game is loaded'
		
		# Set up mod-specific values
		self.initValues()

	def onEndGameTurn(self, argsList):
		'Called at the end of the end of each turn'
		iGameTurn = argsList[0]
		
		# Everything that happens at the end of a turn pertaining to the mod comes through here
		self.turnChecker(iGameTurn)
		
	def onCityBuilt(self, argsList):
		'City Built'
		city = argsList[0]
		
		# Init city script data (unit spawn counter)
		self.initCity(city)
		
	def onCityAcquired(self, argsList):
		'City Acquired'
		
		owner,playerType,city,bConquest,bTrade = argsList

		if( not bConquest ) :
                    # Init city script data (unit spawn counter)
                    self.initCity(city)
                    return

		newOwnerID = city.getOwner()
		oldOwnerID = city.getPreviousOwner()
		orgOwnerID = city.getOriginalOwner()		

		if( newOwnerID == gc.getBARBARIAN_PLAYER() ) : #and self.getCitySpawnUnitCounter( oldOwnerID, city ) > 0 ) :
                    if( self.LOG_DEBUG ) : print("Rebel - City %s captured by barb rebels!"%(city.getName()))
                    oldOwner = gc.getPlayer(oldOwnerID)
                    
                    if( not oldOwnerID == orgOwnerID ) :
                        orgOwner = gc.getPlayer(orgOwnerID)
                        if( self.LOG_DEBUG ) : print("Rebel - City %s originally owner by %s!"%(city.getName(),orgOwner.getCivilizationDescription(0)))

                        #if( something ) :
                            # Flip back to original owner

                            # Kill nearby barbs                    

                    if( city.countTotalCulture() > 100 ) :
                        if( not (oldOwnerID == city.findHighestCulture()) ) :
                            cultOwner = gc.getPlayer( city.findHighestCulture() )
                            if( self.LOG_DEBUG ) : print("Rebel - City %s culturally dominated by %s!"%(city.getName(),cultOwner.getCivilizationDescription(0)))

                            #if( something ) :
                                # Flip to cultural owner

                                # Gift nearby barbs

                elif( gc.getPlayer( newOwnerID ).getNumCities() <= 1 ) :
                    if( self.LOG_DEBUG ) : print("Rebel - City %s captured by rebel player %d!!!"%(city.getName(),newOwnerID))
                    
                    city.setCulture( newOwnerID, city.countTotalCulture(), True )
                    ix = city.getX()
                    iy = city.getY()

                    newOwner = gc.getPlayer( newOwnerID )

                    if( city.isVisible( game.getActiveTeam(), False ) ) :
                        CyInterface().addImmediateMessage("The rebel %s captures its first city, %s!  Support for the rebellion swells ..."%(newOwner.getCivilizationDescription(0),city.getName()),"")

                    iBestDefender = UnitTypes.NO_UNIT
                    iCounter = UnitTypes.NO_UNIT
                    iAttack = UnitTypes.NO_UNIT
                    for unitID in range(0,gc.getNumUnitInfos()) :
                        if( city.canTrain(unitID,False,False) and gc.getUnitInfo(unitID).getDomainType() == DomainTypes.DOMAIN_LAND ): # canTrain only does technology check
                            unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                            # Defender (Archer,Longbow)
                            if( gc.getUnitInfo(unitID).getDefaultUnitAIType() == UnitAITypes.UNITAI_CITY_DEFENSE ):
                                if( (iBestDefender == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() >= gc.getUnitInfo(iBestDefender).getCombat() ) :
                                    #if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Civ def units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                    if( unitID == gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                        iBestDefender = unitID
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Best defender set to %s"%(PyInfo.UnitInfo(iBestDefender).getDescription()))
                            # Counter (Axemen,Phalanx)
                            if( gc.getUnitInfo(unitID).getDefaultUnitAIType() == UnitAITypes.UNITAI_COUNTER ):
                                if( (iCounter == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() >= gc.getUnitInfo(iCounter).getCombat() ) :
                                    #if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Civ counter units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                    if( unitID == gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                        iCounter = unitID
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Best counter unit set to %s"%(PyInfo.UnitInfo(iCounter).getDescription()))
                            # Assault units
                            if( gc.getUnitInfo(unitID).getUnitAIType( UnitAITypes.UNITAI_ATTACK ) ):
                                if( (iAttack == UnitTypes.NO_UNIT) or gc.getUnitInfo(unitID).getCombat() > gc.getUnitInfo(iAttack).getCombat() ) :
                                    #if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Civ attack units return value: %d from %d, comp to %d"%(gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass),unitClass,unitID))
                                    if( unitID == gc.getCivilizationInfo( newOwner.getCivilizationType() ).getCivilizationUnits(unitClass) ) :
                                        iAttack = unitID
                                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("  BC - Best attack set to %s"%(PyInfo.UnitInfo(iAttack).getDescription()))

                    if( not iBestDefender == UnitTypes.NO_UNIT ) : newOwner.initUnit( iBestDefender, ix, iy, UnitAITypes.NO_UNITAI )
                    if( not iBestDefender == UnitTypes.NO_UNIT ) : newOwner.initUnit( iBestDefender, ix, iy, UnitAITypes.NO_UNITAI )
                    if( not iCounter == UnitTypes.NO_UNIT ) : newOwner.initUnit( iCounter, ix, iy, UnitAITypes.NO_UNITAI )
                    if( not iAttack == UnitTypes.NO_UNIT ) : newOwner.initUnit( iAttack, ix, iy, UnitAITypes.NO_UNITAI )
		
		# Init city script data (unit spawn counter)
		self.initCity(city)

###########################################################################################
####################################### REBEL EVENTS ######################################
###########################################################################################
	
	def checkForRebellion(self, iTurn):
		
		# Loop through all players...
		for iPlayerID in range(gc.getMAX_CIV_PLAYERS()):
			
			pPlayer = gc.getPlayer(iPlayerID)
#			pyPlayer = PyPlayer(iPlayerID)
			
			# If player is alive...
			if (pPlayer.isAlive()):
				
########################## This is where the fun begins ##########################
				
				# This player's state religion
				iStateReligionID = pPlayer.getStateReligion()
				
				# Civics
				iCivicsValue = self.getCivicsFactor(pPlayer)
				
				# Civ size in land plots
				iCivNumLandPlots = pPlayer.getTotalLand()
				
########################## Loop through player's cities... ##########################
				
				apCityList = self.getCityList(pPlayer)
				for pCity in apCityList:

                                    unhappyPop = (pCity.unhappyLevel(0) - pCity.happyLevel())

                                    if( pCity.isOccupation() ) :
                                        if( self.LOG_DEBUG ) : print("Rebel - City %s is under occupation by %s" %(pCity.getName(),pPlayer.getCivilizationDescription(0)))
                                    elif( unhappyPop > 0 or self.allowHappyRebellions ) :
					
					# City factors
					
					# City religions vs. state religion
					iReligionValue = self.getReligionFactor(iStateReligionID, pCity)
					# Modify religion value based on civics
					iCivicOptionReligion = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_RELIGION')
                                        iReligionCivic = pPlayer.getCivics(iCivicOptionReligion)
                                        iOrganizedReligionID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_ORGANIZED_RELIGION')
                                        iTheocracyID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_THEOCRACY')
                                        if( iReligionCivic == iOrganizedReligionID ) :
                                            iReligionValue = iReligionValue*self.iOrganizedReligionMultiplier
                                        elif( iReligionCivic == iTheocracyID ) :
                                            iReligionValue = iReligionValue*self.iTheocracyMultiplier
					# Make religion less important after the game is nearing completion
					iHalfTurnNum = (CyGame().getMaxTurns() - CyGame().getStartTurn()) / 2
					if (CyGame().getGameTurn() > iHalfTurnNum):
						iTurnsAfterHalf = CyGame().getGameTurn() - iHalfTurnNum
						fReligionMod = iTurnsAfterHalf / (iHalfTurnNum * 1.0)
						iReligionValue -= iReligionValue * fReligionMod
					
					# Health
					iHealthValue = (pCity.badHealth(false) - pCity.goodHealth()) * self.iHealthFactor
					if( iHealthValue < 0 ) :
                                            iHealthValue = 0
										
					# Happiness
					iHappinessValue = (pCity.unhappyLevel(0) - pCity.happyLevel()) * self.iHappinessFactor
					if (iHappinessValue < 0):
                                                # Happy people will over look SOME of your other problems
						iHappinessValue = iHappinessValue/self.iHappyDivisor
					
					# City nationality
					iNationalityValue = (100 - pCity.plot().calculateCulturePercent(iPlayerID)) * self.iNationalityFactor / 100.0
					if( pCity.plot().countTotalCulture() < 100 ) :
                                                # If no civ has much culture here, reduce factor
                                                iNationalityValue = int(iNationalityValue*(pCity.countTotalCulture()/100.0))

##                                        if( iNationalityValue > 0 ) :
##                                                print "Rebel - Culture percent in %s: %d, plot: %d"%(pCity.getName(),pCity.calculateCulturePercent(iPlayerID),pCity.plot().calculateCulturePercent(iPlayerID))
##					print("XXX - Player %d Culture Percent: %d; (100 - %d) * %d" %(iPlayerID, pCity.calculateCulturePercent(iPlayerID), pCity.calculateCulturePercent(iPlayerID), self.iNationalityFactor))
					
					# Civ size
					defNumPlayers = gc.getWorldInfo(gc.getMap().getWorldSize()).getDefaultPlayers()
					expCivPlots = CyMap().getLandPlots() / ( defNumPlayers * 1.0)
					iCivSizeValue = iCivNumLandPlots / expCivPlots * self.iCivSizeFactor
                                        pCapital = pPlayer.getCapitalCity()
                                        # Distance to capital divided by something a little bigger than civ's "radius"
                                        cityDistModifier = plotDistance( pCity.getX(), pCity.getY(), pCapital.getX(), pCapital.getY() )/(sqrt(iCivNumLandPlots/3.0))
                                        oCityDistMod = cityDistModifier
                                        cityDistModifier = min([cityDistModifier, self.maxCityDistModifier])
                                        #if( self.LOG_DEBUG ) : print("Rebel - City %s has dist mod %f from %f"%(pCity.getName(),cityDistModifier,oCityDistMod))
                                        # Total civ size effect: some effect for all cities, more for distant cities
                                        iCivSizeValue = int( self.fLandModifier*(iCivSizeValue/2.0 + cityDistModifier*iCivSizeValue/2.0) )

					# City area
					iCityAreaValue = 0
					if( not (pCity.area().getID() == pPlayer.getCapitalCity().area().getID()) ) :
                                                iCityAreaValue += self.iCityAreaFactor
					
					# Sum all factors
					iTotalRebelValue = iReligionValue + iCivicsValue + iHealthValue + iHappinessValue + iNationalityValue + iCivSizeValue
                                        #if( self.LOG_DEBUG ) : print("XXX - City %s has iTotalRebelValue of %d - Civics %d, Religion %d, Health %d, Happiness %d, Nationality %d, CivSize %d, CityArea %d" %(pCity.getName(), iTotalRebelValue, iCivicsValue, iReligionValue, iHealthValue, iHappinessValue, iNationalityValue, iCivSizeValue, iCityAreaValue))
                                        iTotalRebelValue = int( iTotalRebelValue )
				
                                        if( iTotalRebelValue > self.iThreshold ) :
                                            if( self.LOG_DEBUG ) : print("Rebel - City %s has iTotalRebelValue of %d - Civics %d, Religion %d, Health %d, Happiness %d, Nationality %d, CivSize %d, CityArea %d" %(pCity.getName(), iTotalRebelValue, iCivicsValue, iReligionValue, iHealthValue, iHappinessValue, iNationalityValue, iCivSizeValue, iCityAreaValue))
					
                                            # Now spawn units
                                            self.doRebellion(iPlayerID, pCity, iTotalRebelValue)
	
	def getReligionFactor(self, iStateReligionID, pCity):
		
		iReligionFactor = 0
		
		# Loop through all religions
		for iCityReligionID in range(gc.getNumReligionInfos()):
			
			# If city has this religion...
			if (pCity.isHasReligion(iCityReligionID)):
				
				# If this city religion doesn't match the state religion...
				if (iStateReligionID != iCityReligionID):
					
					iReligionFactor += self.iDifferentReligionFactor
				else :
                                        iReligionFactor -= self.iDifferentReligionFactor/2
		
		return iReligionFactor
	
	def getCivicsFactor(self, pPlayer):
		
		# Government Civics
		iCivicOptionGovernment = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_GOVERNMENT')
		iGovernmentCivic = pPlayer.getCivics(iCivicOptionGovernment)
		
		iDespotismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_DESPOTISM')
		iHereditaryRuleID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_HEREDITARY_RULE')
		iPoliceStateID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_POLICE_STATE')
		iUniversalSuffrageID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_UNIVERSAL_SUFFRAGE')
		
		# Legal Civics
		iCivicOptionLegal = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_LEGAL')
		iLegalCivic = pPlayer.getCivics(iCivicOptionLegal)
		
		iBarbarismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_BARBARISM')
		iVassalageID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_VASSALAGE')
		
		# Labor Civics
		iCivicOptionLabor = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_LABOR')
		iLaborCivic = pPlayer.getCivics(iCivicOptionLabor)
		
		iTribalismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_TRIBALISM')
		iSlaveryID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_SLAVERY')
		iSerfdomID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_SERFDOM')
		iCasteSystemID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_CASTE_SYSTEM')
		
		# Religion Civics
		iCivicOptionReligion = CvUtil.findInfoTypeNum(gc.getCivicOptionInfo, gc.getNumCivicOptionInfos(), 'CIVICOPTION_RELIGION')
		iReligionCivic = pPlayer.getCivics(iCivicOptionReligion)
		
		iPaganismID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_PAGANISM')
		iOrganizedReligionID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_ORGANIZED_RELIGION')
		iTheocracyID = CvUtil.findInfoTypeNum(gc.getCivicInfo, gc.getNumCivicInfos(), 'CIVIC_THEOCRACY')
		
		########## CALCULATE VALUES ##########
		
		iCivicValue = 0
		
		if (iGovernmentCivic == iDespotismID):
			iCivicValue += self.iDespotismFactor
		elif (iGovernmentCivic == iHereditaryRuleID):
			iCivicValue += self.iHereditaryRuleFactor
		elif (iGovernmentCivic == iPoliceStateID):
			iCivicValue += self.iPoliceStateFactor
		elif ( iGovernmentCivic ==  iUniversalSuffrageID ) :
                        iCivicValue += self.iUniversalSuffrageFactor
			
		if (iLegalCivic == iBarbarismID):
			iCivicValue += self.iBarbarismFactor
		elif (iLegalCivic == iVassalageID):
			iCivicValue += self.iVassalageFactor
			
		if (iLaborCivic == iTribalismID):
			iCivicValue += self.iTribalismFactor
		elif (iLaborCivic == iSlaveryID):
			iCivicValue += self.iSlaveryFactor
		elif (iLaborCivic == iSerfdomID):
			iCivicValue += self.iSerfdomFactor
		elif (iLaborCivic == iCasteSystemID):
			iCivicValue += self.iCasteSystemFactor
			
		if (iReligionCivic == iPaganismID):
			iCivicValue += self.iPaganismFactor
		elif (iReligionCivic == iOrganizedReligionID):
			iCivicValue += self.iOrganizedReligionFactor
		elif (iReligionCivic == iTheocracyID):
			iCivicValue += self.iTheocracyFactor
		
		return iCivicValue

###########################################################################################
####################################### OTHER EVENTS ######################################
###########################################################################################
	
	def doRebellion(self, iPlayerID, pCity, iTotalRebelValue):

            iCityX = pCity.getX()
	    iCityY = pCity.getY()
		
            # If this city has spawned a unit recently then no more units
            if (self.getCitySpawnUnitCounter(iPlayerID, pCity) == 0):
                    
                iUnitSpawnRoll = self.getRand(self.iRebelSpawnRoll)

                if (iUnitSpawnRoll < int(self.fModifier*iTotalRebelValue)):
                    # City is rebelling!

                    self.rebelOffer( iPlayerID, pCity, iTotalRebelValue )

                    #self.spawnRebels( iPlayerID, pCity, iTotalRebelValue )

            else :
                if( self.LOG_DEBUG ) : print ("Rebel - City %s just revolted"%(pCity.getName()))

        

        def rebelOffer( self, iPlayerID, pCity, iTotalRebelValue ) :

            pPlayer = gc.getPlayer( iPlayerID )
            iEra = pPlayer.getCurrentEra()
            iRandCost = 800 - self.getRand( 1000 )
            iBuyoffCost = int( (iTotalRebelValue + iRandCost)/(3+gc.getNumEraInfos()-iEra) )

            if( pPlayer.isHuman() ) :
                self.iBuyoffCost = iBuyoffCost
                self.pCity = pCity
                self.iTotalRebelValue = iTotalRebelValue
                #CvCameraControls.centerCameraByObject( pCity )
                CyCamera().JustLookAt( pCity.plot().getPoint() )
                
                # Offer human player the chance to negotiate with Rebels
                if( self.LOG_DEBUG ) : print ("Rebel - Buy off cost for this rebellion is %d"%(self.iBuyoffCost))

                popup = PyPopup.PyPopup(self.rebelOfferPopup,contextType = EventContextTypes.EVENTCONTEXT_ALL)
                popup.setHeaderString( 'A Rebellion is Brewing!' )
                bodStr = 'A leader from the the rebellious city of %s approaches you with an offer.\n\n'%(pCity.getName())
                bodStr += 'They would be willing to put down their arms and wait to see if conditions improve for the sum of %d.'%(self.iBuyoffCost)
                if( pPlayer.getGold() < self.iBuyoffCost ) :
                    bodStr += '  However, you do not have enough in the treasury.'
                popup.setBodyString( bodStr )
                popup.addSeparator()

                popup.addButton('The %s does not negotiate with rebels.'%(pPlayer.getCivilizationDescription(0)))
                popup.addButton('We can afford the %d gold.'%(self.iBuyoffCost))
                #popup.addPythonButton( 'Rebellion.doBuyoff', 'We can afford the %d gold.'%(self.iBuyoffCost), 'Will buyoff the rebels, putting off the rebellion for some time', iData1 = iPlayerID, iData2 = pCity.getID() )

                popup.launch(bCreateOkButton = False)
                
                return

            else :
                # AI player's city is rebelling, see if they can buy them off
                iBuyoffCost = iBuyoffCost/2
                if( self.LOG_DEBUG ) : print ("Rebel - AI Buy off cost for this rebellion is %d"%(iBuyoffCost))
                
                if( pPlayer.getGold() >= iBuyoffCost ) :
                    iAgg = CvUtil.findInfoTypeNum(gc.getTraitInfo,gc.getNumTraitInfos(),'TRAIT_AGGRESSIVE')
                    if( pPlayer.hasTrait( iAgg ) ) :
                        base = .85
                    else :
                        base = .7

                    oddsForBuy = 1.0 - pow( base, pPlayer.getGold()/(1.0*iBuyoffCost) )
                    oddsForBuy = int(100*oddsForBuy)

                    if( self.LOG_DEBUG ) : print ("Rebel - Odds for buyoff are %d, with base %f"%(oddsForBuy,base))

                    if( oddsForBuy >= self.getRand( 100 ) ) :
                        if( self.LOG_DEBUG ) : print ("Rebel - AI is buying off rebels")
                        pPlayer.changeGold( -iBuyoffCost )
                        self.setCitySpawnUnitCounter(iPlayerID, pCity, self.iCitySpawnUnitCounterMax)
                    else :
                        if( self.LOG_DEBUG ) : print ("Rebel - AI opted not to buy off rebels")
                        self.spawnRebels( iPlayerID, pCity, iTotalRebelValue )
                    
                else :
                    if( self.LOG_DEBUG ) : print ("Rebel - AI doesn't have the money")
                    self.spawnRebels( iPlayerID, pCity, iTotalRebelValue )

##        def doBuyoff( self, iPlayerID, iCityID ) :
##            print ("Rebel - Got to do Buyoff!!!")
##
##            pPlayer = gc.getPlayer( iPlayerID )
##
##            pCity = pPlayer.getCity( iCityID )
##
##            gc.getPlayer( iPlayerID ).changeGold( self.iBuyoffCost )
##
##            self.setCitySpawnUnitCounter( iPlayerID, pCity, self.iCitySpawnUnitCounterMax)

        def rebelOfferHandler( self, iPlayerID, netUserData, popupReturn ) :
           
            if( popupReturn.getButtonClicked() == 0 ) :
                if( self.LOG_DEBUG ) : print ("Rebel - Clicked option 0, no deal")

                self.spawnRebels( iPlayerID, self.pCity, self.iTotalRebelValue )

            elif( popupReturn.getButtonClicked() == 1 ) :
                if( self.LOG_DEBUG ) : print ("Rebel - Clicked option 1, buyoff")

                gc.getPlayer( iPlayerID ).changeGold( -self.iBuyoffCost )

                self.setCitySpawnUnitCounter( iPlayerID, self.pCity, self.iCitySpawnUnitCounterMax)

            else :
                print ("Rebel - Unexpected return from rebel offer popup")
                self.spawnRebels( iPlayerID, self.pCity, self.iTotalRebelValue )
					
	def spawnRebels(self, iPlayerID, pCity, iTotalRebelValue):

                if( pCity.isVisible( game.getActiveTeam(), False ) ) :
                    CyInterface().addImmediateMessage("A Rebellion has occured in %s!!!"%(pCity.getName()),"")
		
		iCityX = pCity.getX()
		iCityY = pCity.getY()
					
                aiSpawnPlot = self.findUnitPlacementPlot(self.iBarbariansID, iCityX, iCityY, true, true, self.iUnitSpawnRange)
#				print("Spawn Plot: %d, %d" %(aiSpawnPlot[0], aiSpawnPlot[1]))
                if (aiSpawnPlot == "Oh Snap"):
                    self.setCitySpawnUnitCounter(iPlayerID, pCity, self.iCitySpawnUnitCounterMax/2)
                    if( self.LOG_DEBUG ) : print("Rebel:  No spawnable plot!!!  Aborting Rebellion")
                    return
                else:
                        pRebel = None
                        
                        if( self.rebelsAsBarbs ) :
                            pRebel = gc.getPlayer(self.iBarbariansID)
                        else :
                            # Is there another rebellion we can join?
                            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                                if( gc.getPlayer(i).getNumUnits() > 0 ) :
                                    teami = gc.getTeam( gc.getPlayer(i).getTeam() )
                                    owner = gc.getPlayer( pCity.getOwner() )
                                    if( teami.isAtWar(owner.getTeam()) ) :
                                        if( self.LOG_DEBUG ) : print("Rebel:  Player %d is at war with owner of %s"%(i,pCity.getName()))
                                        if( not gc.getPlayer(i).isFoundedFirstCity() ) :  # Is a homeless rebel
                                            if( self.LOG_DEBUG ) : print("Rebel:  Player %d (homeless rebel) is rebelling against owner of %s!!!  Joining rebellion."%(i,pCity.getName()))
                                            pRebel = gc.getPlayer(i)
                                            break
                                        elif( gc.getPlayer(i).getNumCities() < 4 ) :
                                            if( game.getGameTurn() - gc.getPlayer(i).getCapitalCity().getGameTurnAcquired() < 30 ) :
                                                if( self.LOG_DEBUG ) : print("Rebel:  Player %d (young rebel) is rebelling against owner of %s!!!  Joining rebellion."%(i,pCity.getName()))
                                                pRebel = gc.getPlayer(i)
                                                break

                            if( pRebel == None ) :
                                # Is there a dead civ in this city that can rise from the ashes?
                                for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                                    if( gc.getPlayer(i).isEverAlive() and (gc.getPlayer(i).getNumUnits() == 0) ) :
                                        if( pCity.getCulture( i ) > 10 ) :
                                            if( self.LOG_DEBUG ) : print("Rebel:  Dead player %d has culture in rebelling %s, respawning"%(i,pCity.getName()))
                                            pRebel = gc.getPlayer(i)
                                            break

                            if( pRebel == None ) :
                                # Create a new rebellion (or return barb if not possible)
                                pRebel = self.spawnRebelPlayer( pCity )

                        bIsBarb = (pRebel.getID() == self.iBarbariansID)

                        spawnableUnits = list()

                        for unitID in range(0,gc.getNumUnitInfos()) :
                            if( pCity.canTrain(unitID,False,False) and gc.getUnitInfo(unitID).getDomainType() == DomainTypes.DOMAIN_LAND ):
                                if( gc.getUnitInfo(unitID).getUnitAIType( UnitAITypes.UNITAI_ATTACK ) ):
                                    unitClass = gc.getUnitInfo(unitID).getUnitClassType()
                                    if( bIsBarb ) :
                                        domPlayerIdx = pCity.plot().calculateCulturalOwner()
                                        domPlayer = gc.getPlayer(domPlayerIdx)
                                        spawnUnitID = gc.getCivilizationInfo( domPlayer.getCivilizationType() ).getCivilizationUnits(unitClass)
                                    else :
                                        spawnUnitID = gc.getCivilizationInfo( pRebel.getCivilizationType() ).getCivilizationUnits(unitClass)
                                        
                                    if( not spawnUnitID in spawnableUnits ) :
                                        spawnableUnits.append( spawnUnitID )
                                        if( gc.getUnitInfo(unitID).getCombat() > 4 ) :
                                            spawnableUnits.append( spawnUnitID )
                                            if( gc.getUnitInfo(unitID).getCombat() > 15 ) :
                                                spawnableUnits.append( spawnUnitID )
                                        if( self.LOG_DEBUG ) : print "Rebel:  Can spawn %s"%(PyInfo.UnitInfo(spawnUnitID).getDescription())

                        if( len(spawnableUnits) < 1 ) :
                            if( self.LOG_DEBUG ) : print("Rebel:  No spawnable units!!!  Aborting Rebellion")
                            return

                        numUnits = int( self.fSizeModifier*(iTotalRebelValue + self.getRand(1500))/(1.0*self.iThreshold) ) + 1

                        for i in range(0,numUnits) :
                            newUnitID = spawnableUnits[self.getRand( len(spawnableUnits) )]
                                                                
                            if( self.LOG_DEBUG ) : print ("Rebel - Player %d in revolution; Rebel %s spawning at [%d, %d]" %(iPlayerID,PyInfo.UnitInfo(newUnitID).getDescription(), aiSpawnPlot[0], aiSpawnPlot[1]))
                            pRebel.initUnit(newUnitID, aiSpawnPlot[0], aiSpawnPlot[1], UnitAITypes.UNITAI_ATTACK)

                        if( not pRebel.getID() == self.iBarbariansID ) :
                            gc.getTeam( pRebel.getTeam() ).declareWar( pCity.getOwner(), True )
                            if( self.LOG_DEBUG ) : print ("Rebel - Rebel civ %s has declared war on %s" %(pRebel.getCivilizationDescription(0),gc.getPlayer(pCity.getOwner()).getCivilizationDescription(0)))
                        
                        self.setCitySpawnUnitCounter(iPlayerID, pCity, self.iCitySpawnUnitCounterMax)

        def spawnRebelPlayer( self, pCity, giveTechs = True ) :
            
            owner = gc.getPlayer( pCity.getOwner() )

            if( owner.getID() == self.iBarbariansID ) :
                return gc.getPlayer(self.iBarbariansID)
            
            cultOwner = None
            if( pCity.countTotalCulture() > 0 ) :
                cultOwner = gc.getPlayer( pCity.findHighestCulture() )
                if( cultOwner.getID() == self.iBarbariansID ) :
                    cultOwner = None

            if( cultOwner == None ) :
                cultOwner = owner

            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Spawning a rebel civ for the %s"%(cultOwner.getCivilizationDescription(0)))

            newPlayerIdx = -1
            # Overwrite a civ that never managed to found a city (probably a previous rebel)
            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                if( (not gc.getPlayer(i).isAlive()) and (not gc.getPlayer(i).isFoundedFirstCity()) and (gc.getPlayer(i).getNumUnits() == 0) ) :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Dead civ %d has founded city? %d"%(i,gc.getPlayer(i).isFoundedFirstCity()))
                    newPlayerIdx = i
                    break

            if( newPlayerIdx < 0 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - No 'fake' civs, looking for empty slots")
                # Put in empty slot
                for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                    if( (not gc.getPlayer(i).isEverAlive()) and (gc.getPlayer(i).getNumUnits() == 0) ) :
                        newPlayerIdx = i
                        break

            if( newPlayerIdx < 0 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - No available slots (%d), spawning as Barbarians"%(newPlayerIdx))
                return gc.getPlayer(self.iBarbariansID)

            # Don't reincarnate as either of these
            iMinor = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_MINOR')
            iBarbarian = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_BARBARIAN')
            # Civs not currently in the game
            availableCivs = list()
            # Civs with similar style to cultOwner, if they exist
            similarStyleCivs = list()
            for civType in range(0,gc.getNumCivilizationInfos()) :
                if( not civType == iBarbarian ) :
                    if( not civType == iMinor ) :
                        taken = False
                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                            if( civType == gc.getPlayer(i).getCivilizationType() ) :
                                if( gc.getPlayer(i).isAlive() or gc.getPlayer(i).isFoundedFirstCity() ) :
                                    taken = True
                                    break
                        if( not taken ) :
                            availableCivs.append(civType)
                            if( not cultOwner == None ) :
                                if( gc.getCivilizationInfo( cultOwner.getCivilizationType() ).getArtStyleType() == gc.getCivilizationInfo(civType).getArtStyleType() ) :
                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Potential similar style civ: %s"%(gc.getCivilizationInfo(civType).getShortDescription(0)))
                                    similarStyleCivs.append(civType)

            if( len(availableCivs) < 1 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - No available civs, spawning as Barbarians")
                return gc.getPlayer(self.iBarbariansID)

            if( len(similarStyleCivs) > 0 ) :
                # Create a civ using this style
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Creating a similar style civ to %s"%(cultOwner.getCivilizationDescription(0)))
                availableCivs = similarStyleCivs

            newCivIdx = availableCivs[game.getSorenRandNum(len(availableCivs),'Rebel: pick unused civ type')]

            leaderList = list()
            for leaderType in range(0,gc.getNumLeaderHeadInfos()) :
                if( gc.getCivilizationInfo(newCivIdx).isLeaders(leaderType) ) :
                    leaderList.append(leaderType)

            if( len(leaderList) < 1 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Unexpected lack of possible leaders, spawning as Barbarians")
                return gc.getPlayer(self.iBarbariansID)
            
            newLeaderIdx = leaderList[game.getSorenRandNum(len(leaderList),'Rebel: pick leader')]

            success = game.createNewPlayer( newPlayerIdx, newCivIdx, newLeaderIdx, -1, False, 1, 1 )

            if( not success ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - New civ creation failed, spawning as Barbarians")
                return gc.getPlayer(self.iBarbariansID)

            pRebel = gc.getPlayer(newPlayerIdx)

            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Rebel civ %s spawned in slot %d"%(pRebel.getCivilizationDescription(0),newPlayerIdx))

            if( giveTechs ) :
                # Give all techs known by city owner, except a few of the most expensive
                knownTechs = list()
                mostExpensive = list()
                minMostExpensive = 0
                numMostExpensive = 2 + self.getRand(4)

                ownerTeam = gc.getTeam( owner.getID() )
                
                for techID in range(0,gc.getNumTechInfos()) :
                    if( ownerTeam.isHasTech( techID ) ) :
                        knownTechs.append( techID )
                        if( gc.getTechInfo( techID ).getResearchCost() > minMostExpensive ) :
                            if( len(mostExpensive) < numMostExpensive ) :
                                mostExpensive.append( techID )
                            else :
                                for j in range(0,len(mostExpensive)) :
                                    if( gc.getTechInfo( mostExpensive[j] ).getResearchCost() == minMostExpensive ) :
                                        mostExpensive.remove( mostExpensive[j] )
                                        break
                                mostExpensive.append( techID )
                                minMostExpensive = gc.getTechInfo( mostExpensive[0] ).getResearchCost()
                                for j in range(0,len(mostExpensive)) :
                                    if( gc.getTechInfo( mostExpensive[j] ).getResearchCost() < minMostExpensive ) :
                                        minMostExpensive = gc.getTechInfo( mostExpensive[j] ).getResearchCost()

                rebelTeam = gc.getTeam( pRebel.getID() )
                for techID in knownTechs :
                    if( not techID in mostExpensive ) :
                        rebelTeam.setHasTech(techID,True,pRebel.getID(),False,False)
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Giving rebel %s"%(PyInfo.TechnologyInfo(techID).getDescription()))
                    else :
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - NOT giving rebel %s"%(PyInfo.TechnologyInfo(techID).getDescription()))

            return pRebel

        def removeFloatingRebellions(self) :
            # Destroy all units for a rebel who is at peace, yet has no cities

            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                player = gc.getPlayer(i)
                if( not player.isFoundedFirstCity() and player.getNumUnits() > 0 ) :
                    if( player.getNumUnits() == player.getNumMilitaryUnits() ) :
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Player %d is a homeless rebel"%(player.getID()))
                        if( gc.getTeam( player.getTeam() ).getAtWarCount( True ) == 0 ) :
                            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Rebel - Player %d has lost thier cause, terminating rebel"%(player.getID()))
                            player.killUnits()
	
	def alterCitySpawnUnitCounters(self):
		
		# Loop through all players...
		for iPlayerLoop in range(gc.getMAX_CIV_PLAYERS()):
			
			pPlayer = gc.getPlayer(iPlayerLoop)
			
			# If this player is alive...
			if (pPlayer.isAlive()):
				
				apCityList = self.getCityList(pPlayer)
				
				# Loop through all this player's cities...
				for pCity in apCityList:
					
					iCityID = pCity.getID()
					# If this city's spawn counter is greater than 0...
					if (self.getCitySpawnUnitCounter(iPlayerLoop, pCity) > 0):
						# Reduce CitySpawnUnit counter by 1 in applicable cities
						self.changeCitySpawnUnitCounter(iPlayerLoop, iCityID, -1)

###########################################################################################
#################################### UTILITY FUNCTIONS ####################################
###########################################################################################
		
	def getRand(self, iNum):
		
		return CyGame().getSorenRandNum(iNum, "RebellionModScenario")
		
	def findUnitPlacementPlot(self, iPlayerID, iPlotX, iPlotY, bLand = true, bIncludePlot = true, bSameArea = True, iRange = 3):
		
		aTHEPlot = []
		aiPossiblePlots = []
		iNumPossiblePlots = 0
		aiTempPlotHolder = []
		
		pThisPlayer = gc.getPlayer(iPlayerID)
		pThisTeam = gc.getTeam(pThisPlayer.getTeam())

		iBaseArea = CyMap().plot(iPlotX, iPlotY).area().getID()
		
		# Loop through plots in usable range
		for iXLoop in range(iPlotX - iRange, iPlotX + iRange):
			for iYLoop in range(iPlotY - iRange, iPlotY + iRange):
				
				# Map bounds
				if (iXLoop >= 0 and iXLoop < CyMap().getGridWidth() and iYLoop >= 0 and iYLoop < CyMap().getGridHeight()):
					
					# Check to see if center plot should be excluded
					if (bIncludePlot or (not bIncludePlot and (iXLoop != iPlotX) and (iYLoop != iPlotY))):
						
						pPlot = CyMap().plot(iXLoop, iYLoop)
						
						# Don't include enemy city plots!
						if (pPlot.isEnemyCity(pThisPlayer.getTeam()) == false):
							
							# Passable?
							if (not pPlot.isImpassable()):
								
								# Check whether this plot type matches the type requested
								if ((bLand == true and CyMap().plot(iXLoop, iYLoop).isWater() == false) or \
								    (bLand == false and CyMap().plot(iXLoop, iYLoop).isWater() == true)):

                                                                    if( (not bSameArea) or (CyMap().plot(iXLoop, iYLoop).area().getID() == iBaseArea) ) :
									aiTempPlotHolder = [iXLoop, iYLoop]
									aiPossiblePlots.append(aiTempPlotHolder)
									iNumPossiblePlots += 1
					
		if (iNumPossiblePlots == 0):
			return "Oh Snap"
		
		# Now that our list of possible plots is made, we pick one of them randomly and shove aside any enemy units already there
		iChosenPlot = self.getRand(iNumPossiblePlots - 1)
		aTHEPlot = aiPossiblePlots[iChosenPlot]
		
		# Check to see if there are any other units already on this plot
		pPlot = CyMap().plot(aTHEPlot[0], aTHEPlot[1])
		
		if (pPlot.getNumUnits() > 0):
			
			bEnemiesPresent = true
			
			while(pPlot.getNumUnits() > 0 and bEnemiesPresent):
				pLoopUnit = pPlot.getUnit(0)
				pUnitOwner = gc.getPlayer(pLoopUnit.getOwner())
				
				if (pUnitOwner):
					# Check to see if this unit belongs to a player that 'our' player is at war with
					if (pThisTeam.isAtWar(pUnitOwner.getTeam())):
						# Move this enemy unit
						self.moveUnitAside(pLoopUnit)
					else:
						# If it's a friendly, don't move it!
						bEnemiesPresent = false
		
		return aTHEPlot
		
	def moveUnitAside(self, pUnitToMove):
		
		iUnitX = pUnitToMove.getX()
		iUnitY = pUnitToMove.getY()
		
		# Don't look further than 10 plots away
		iMaxRange = 10
		iRing = 1
		
		aaiTestedList = [iUnitX, iUnitY]
		
		pPlayer = gc.getPlayer(pUnitToMove.getOwner())
		pTeam = gc.getTeam(pPlayer.getTeam())
		
		# Loop in rings around the unit's plot until we find another plot to move this unit to
		while (iRing < iMaxRange):
			
			aiTempList = []
			aaiToTestList = []
			iNumToTest = 0
			aaiValidPlotList = []
			iNumValid = 0
			
			# Find plots in this ring range
			for iX in range(-iRing, iRing + 1):
				for iY in range(-iRing, iRing + 1):
					aiTempList = [iX + iUnitX, iY + iUnitY]
					
					# Don't test this plot more than once
					if (aiTempList not in aaiTestedList):
						
						pPlot = CyMap().plot(aiTempList[0], aiTempList[1])
						bValidPlot = false
						
						if (pPlot.isWater() and pUnitToMove.getDomainType() == DomainTypes.DOMAIN_SEA):
							bValidPlot = true
						elif (pPlot.isWater() == false):
							if (pUnitToMove.getDomainType() == DomainTypes.DOMAIN_LAND or pUnitToMove.getDomainType() == DomainTypes.DOMAIN_AIR):
								bValidPlot = true
						
						# Make sure this unit can move to this plot type (no land units on water, etc.)
						if (bValidPlot):
							aaiToTestList.append(aiTempList)
							iNumToTest += 1
					
			# Now test these plots
			for iTestLoop in range(iNumToTest):
				
				pPlot = CyMap().plot(iUnitX, iUnitY)
				
				iX = aaiToTestList[iTestLoop][0]
				iY = aaiToTestList[iTestLoop][1]
				
				# If there are no enemy units here then add to the final valid plot list
				if (not self.isEnemyUnits(pTeam, iX, iY)):
					
					aaiValidPlotList.append(aaiToTestList[iTestLoop])
					iNumValid += 1
					
				# Add this plot to the tested list (to prevent it being looked at again) and look at the next in this ring
				aaiTestedList.append(aaiToTestList[iTestLoop])
				
			# This ring has been iterated through and placed in a final list - now pick a random plot from that list and exit the function
			
			if (iNumValid > 0):
				iChosenPlot = self.getRand(iNumValid - 1)
				
				iX = aaiValidPlotList[iChosenPlot][0]
				iY = aaiValidPlotList[iChosenPlot][1]
				
				pUnitToMove.setXY(iX, iY)
				
				return
				
			# None of the plots in this ring are valid, so time to try the next ring
			
			iRing += 1
		
	def isEnemyUnits(self, pThisTeam, iPlotX, iPlotY):
		
		# Check to see if there are any other units already on this plot
		pPlot = CyMap().plot(iPlotX, iPlotY)
		
		if (pPlot.getNumUnits() > 0):
			
			for iUnitLoop in range(pPlot.getNumUnits()):
				
				pLoopUnit = pPlot.getUnit(iUnitLoop)
				
				pUnitOwner = gc.getPlayer(pLoopUnit.getOwner())
				
				# Check to see if this unit belongs to a player that 'our' player is at war with
				if (pThisTeam.isAtWar(pUnitOwner.getTeam())):
					return true
		
		return false
	
	def getCityList(self, pPlayer):
		iPlayerID = pPlayer.getID()
		
		lCity = []
		(loopCity, iter) = pPlayer.firstCity(false)
		while(loopCity):
			cityOwner = loopCity.getOwner()
			if ( not loopCity.isNone() and loopCity.getOwner() == iPlayerID ): #only valid cities
				pCity = pPlayer.getCity(loopCity.getID())
#				city = PyCity( self.getID(), loopCity.getID() )
				lCity.append(pCity)
			(loopCity, iter) = pPlayer.nextCity(iter, false)
		return lCity
		
	def addPopup(self, szText, szTitle):
		
#		szTitle = self.szGameDate = CyGameTextMgr().getTimeStr(CyGame().getGameTurn(), false)
		
		popup = PyPopup.PyPopup(-1)
		popup.setHeaderString(szTitle)
		popup.setBodyString(szText)
		popup.launch(true, PopupStates.POPUPSTATE_QUEUED)



###########################################################################################
####################################### SCRIPT DATA #######################################
###########################################################################################
	
	def initCity(self, pCity):
		
		# Parameter 0 is City Spawn Unit Counter Max
		szScriptData = [self.iCitySpawnUnitCounterMax]
		
		# Save Script Data - SpawnUnitCounter
		pCity.setScriptData(pickle.dumps(szScriptData))
	
	def getCitySpawnUnitCounter(self, iPlayer, pCity):
		
		pPlayer = gc.getPlayer(iPlayer)
#		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iSpawnUnitCounter = szScriptData[self.iCitySpawnScriptID]
                except EOFError:
                    iSpawnUnitCounter = 0
                    self.setCitySpawnUnitCounter( iPlayer, pCity, 0 )

		#print("Rebel - Loading spawn unit counter, val: %d"%(iSpawnUnitCounter))
		
		return iSpawnUnitCounter
		
	def setCitySpawnUnitCounter(self, iPlayer, pCity, iValue):
		
		pPlayer = gc.getPlayer(iPlayer)
#		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                except EOFError:
                    self.initCity(pCity)
                    szScriptData = pickle.loads(pCity.getScriptData())
		
		# Save Script Data - SpawnUnitCounter
		szScriptData[self.iCitySpawnScriptID] = iValue
		pCity.setScriptData(pickle.dumps(szScriptData))
		
	def changeCitySpawnUnitCounter(self, iPlayer, iCity, iChange):
		
		pPlayer = gc.getPlayer(iPlayer)
		pCity = pPlayer.getCity(iCity)
		
		# Load Script Data - SpawnUnitCounter
		try :
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iSpawnUnitCounter = szScriptData[self.iCitySpawnScriptID]
                except EOFError:
                    self.setCitySpawnUnitCounter( iPlayer, pCity, 0 )
                    szScriptData = pickle.loads(pCity.getScriptData())
                    iSpawnUnitCounter = szScriptData[self.iCitySpawnScriptID]
		
		iSpawnUnitCounter += iChange
		
		# Save Script Data - SpawnUnitCounter
		szScriptData[self.iCitySpawnScriptID] = iSpawnUnitCounter
		pCity.setScriptData(pickle.dumps(szScriptData))
