# REVOLUTION_MOD
#
# by jdog5000
# Version 0.1


from CvPythonExtensions import *
import CvUtil
import PyHelpers		
import Popup as PyPopup

# globals
gc = CyGlobalContext()
PyPlayer = PyHelpers.PyPlayer
PyInfo = PyHelpers.PyInfo
game = CyGame()

class Revolution :

    def __init__(self, customEM, config):

        CvUtil.pyPrint("Initializing Revoltuion Mod")

        self.EventKeyDown=6

        self.pickCivToSplitPopup = 7004
        self.revolutionPopup = 7007

        self.config = config
        self.customEM = customEM

        self.LOG_DEBUG = config.getboolean("Revolution", "DebugMode", False)

        self.minCitiesForSplit = config.getint("Revolution","MinCitiesForSplit",4)
        self.closeRadius = config.getint("Revolution","CloseRadius",10)

        # If the civs capital is lost, their biggest colony may break free
        self.colonySplit = config.getboolean("Revolution", "ColonySplit", False)
        self.minPowerPercentThreshold = config.getint("Revolution", "MinPowerPercent", 20)
        # If the civs capital is lost, the largest held culture block may break free
        self.cultureSplit = config.getboolean("Revolution", "CulturalSplit", False)
        self.minNationalityThreshold = config.getint("Revolution", "MinNationality", 20)
        self.minCultureThreshold = config.getint("Revolution", "MinCulture", 50)
        # If the civs capital is lost, unhappy cities may break free
        self.unhappySplit = config.getboolean("Revolution", "UnhappySplit", False)
        # If the civs capital is lost, their empire may be split in half
        self.empireSplit = config.getboolean("Revolution", "EmpireSplit", False)
        
        if( self.LOG_DEBUG ) : self.customEM.addEventHandler( "kbdEvent", self.onKbdEvent )
        customEM.addEventHandler( "cityLost", self.onCityLost )

        self.customEM.setPopupHandler( self.pickCivToSplitPopup, ["pickCivToSplitPopup",self.pickCivToSplitHandler,self.blankHandler] )
        self.customEM.setPopupHandler( self.revolutionPopup, ["revolutionPopup",self.blankHandler,self.blankHandler] )


    def blankHandler( self, playerID, netUserData, popupReturn ) :
        # Dummy handler to take the second event for popup
        return

    def onKbdEvent(self, argsList ):
        'keypress handler'
        eventType,key,mx,my,px,py = argsList

        if ( eventType == self.EventKeyDown ):
	    theKey=int(key)
	    
            # For debug or trial only
	    if( theKey == int(InputTypes.KB_S) and self.customEM.bShift and self.customEM.bCtrl ) :
                self.pickCivToSplit()
                

    def pickCivToSplit( self ) :
        'Chooser window for forcing split of empire'
        popup = PyPopup.PyPopup(self.pickCivToSplitPopup,contextType = EventContextTypes.EVENTCONTEXT_ALL)
        popup.setHeaderString( 'Split which civ?' )
        popup.addSeparator()
        #popup.createPythonEditBox( '10', 'Number of turns to turn over to AI', 0)
        
        popup.createPythonPullDown( 'Split this civ ...', 1 )
        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
            player = PyPlayer(i)
            if( not player.isNone() ) :
                if( player.isAlive() ) :
                    popup.addPullDownString( "%s of the %s"%(player.getName(),player.getCivilizationName()), i, 1 )

        activePlayerIdx = gc.getActivePlayer().getID()
        popup.popup.setSelectedPulldownID( activePlayerIdx, 1 )
        
        popup.addSeparator()
 
        popup.addButton('Cancel')
        
        popup.launch()

    def pickCivToSplitHandler( self, playerID, netUserData, popupReturn ) :
        'Handles pickCivToSplit popup'
        
        if( popupReturn.getButtonClicked() == 0 ):  # if you pressed cancel
            return
        
        splitIdx = popupReturn.getSelectedPullDownValue( 1 )
        splitPlayer = gc.getPlayer(splitIdx)
        
        if( self.LOG_DEBUG ) : CvUtil.pyPrint( "   CP : You have selected player %d, the %s"%(splitIdx, splitPlayer.getCivilizationDescription(0)) )
        
        wasSplit = self.splitEmpire( splitPlayer, splitPlayer.getCapitalCity(), force = True )


    def onCityLost(self, argsList):
        'City Lost'
        city = argsList[0]
        player = gc.getPlayer(city.getOwner())
        
        if( city.isCapital() ) :
            wasSplit = self.splitEmpire( player, city )

        return            

    def splitEmpire( self, player, capital, force = False ) :
        'Split a civs empire (usually on capital loss'
        
        if( self.LOG_DEBUG and not force ) :
##                popup = PyPopup.PyPopup()
##                bodStr = '%s, the capital of the %s, has been captured!'%(city.getName(), player.getCivilizationName())
##                popup.setBodyString( bodStr )
##                popup.launch()
            print ' Revolt:  %s, the capital of the %s, has been captured!'%(capital.getName(), player.getCivilizationDescription(0))

        playerPy = PyPlayer( player.getID() )
        cityList = playerPy.getCityList()

        if( len(cityList) < self.minCitiesForSplit ) :
            print 'Revolt:  Civ only has %d cities, not splitting'%(len(cityList))
            return False
        else :
            print ' Revolt:  The %s have %d cities'%(player.getCivilizationDescription(0),len(cityList))
        
        # Cities in area with capital
        homelandCities = list()
        # Cities in areas other than capital
        foreignCities = list()
        # Cities with either some built culture, or a percentage nationality
        citiesWithCulture = list()
        # Cities owned originally or previously by another civ
        citiesOtherOwners = list()
        # Cities that are unhappy
        unhappyCities = list()
        
        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
            citiesWithCulture.append(list())
            citiesOtherOwners.append(list())

        # Set up lists of cities for determining how to split empire
        for city in cityList :
            if( not city.isCapital() ) : # Capital has been captured, can't join splitters
                cityCy = city.GetCy()
                if( cityCy.area().getID() == capital.area().getID() ) :
                    homelandCities.append( city )
                else :
                    foreignCities.append( city )

                numUnhappy = city.getUnhappyPopulation() - city.getHappyPopulation()
                if( numUnhappy > 0 ) :
                    unhappyCities.append(city)

                for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                    if( cityCy.plot().calculateCulturePercent(i) > self.minNationalityThreshold ) :
                        citiesWithCulture[i].append(city)
                    elif( cityCy.getCulture(i) > self.minCultureThreshold ) :
                        citiesWithCulture[i].append(city)

                if( (not cityCy.getOwner() == cityCy.getOriginalOwner()) and (not cityCy.getOriginalOwner() == gc.getBARBARIAN_PLAYER()) ) :
                    #if( self.LOG_DEBUG ) : print ' Revolt:  %s originally owned by player %d'%(cityCy.getName(),cityCy.getOriginalOwner())
                    citiesOtherOwners[cityCy.getOriginalOwner()].append(city)
                elif( (not cityCy.getOwner() == cityCy.getPreviousOwner()) and (not cityCy.getPreviousOwner() == PlayerTypes.NO_PLAYER) and (not cityCy.getPreviousOwner() == gc.getBARBARIAN_PLAYER()) ) :
                    #if( self.LOG_DEBUG ) : print ' Revolt:  %s previously owned by player %d'%(cityCy.getName(),cityCy.getPreviousOwner())
                    citiesOtherOwners[cityCy.getPreviousOwner()].append(city)

        if( not len(cityList) == len(homelandCities) + len(foreignCities) + 1 ) :
            if( self.LOG_DEBUG ) : print ' Revolt:  ERROR - something in city listing is wacked out'

        # Begin tests for viability of different splits
        if( self.colonySplit ) : 
            if( self.LOG_DEBUG ) : print ' Revolt:  Attempting colony split'
            if( self.LOG_DEBUG ) : print ' Revolt:  Civ has %d foreign cities'%(len(foreignCities))

            foreignPop = 0
            for city in foreignCities :
                foreignPop += city.getPopulation()

            if( self.LOG_DEBUG ) : print ' Revolt:  Population of foreign cities: %d, of civs capital (highest ever): %d'%(foreignPop,capital.getHighestPopulation())

            # only worth while if there's a large foreign population
            if( foreignPop > capital.getHighestPopulation() ) :
            
                # Build list of cities by area
                areaIDList = list()
                citiesInArea = list()
                popByArea = list()
                powerByArea = list()

                for city in foreignCities :
                    cityCy = city.GetCy()
                    cityArea = cityCy.area().getID()

                    if( cityArea not in areaIDList ) :
                        areaIDList.append(cityArea)
                        citiesInArea.append([city])
                        popByArea.append(city.getPopulation())
                        powerByArea.append(cityCy.area().getPower(player.getID()))
                    else :
                        idx = areaIDList.index(cityArea)
                        citiesInArea[idx].append(city)
                        popByArea[idx] = popByArea[idx] + city.getPopulation()

                # Add in nearby cities
                for city in foreignCities :
                    cityCy = city.GetCy()
                    for idx,areaID in enumerate(areaIDList) :
                        if( not city in citiesInArea[idx] ) :
                            minDist = -1
                            for areaCity in citiesInArea[idx] :
                                dist = plotDistance(city.getX(),city.getY(),areaCity.getX(),areaCity.getY())
                                if( minDist < 0 ) :
                                    minDist =  dist
                                elif( dist < minDist ) :
                                    minDist = dist
                            # Must be within radius and significantly closer to colonies than captial
                            if( minDist > 0 and minDist < self.closeRadius and minDist < .6*plotDistance(city.getX(),city.getY(),capital.getX(),capital.getY() ) ) :
                                citiesInArea[idx].append(city)
                                if( self.LOG_DEBUG ) : print ' Revolt:  %s is close to area %d'%(city.getName(),areaID)
                                popByArea[idx] = popByArea[idx] + city.getPopulation()
                                powerByArea[idx] += cityCy.area().getPower(player.getID())*(city.getPopulation()/cityCy.area().getPopulationPerPlayer(player.getID()))
                        

                if( self.LOG_DEBUG ) : print ' Revolt:  Civ has cities in %d foreign areas'%(len(areaIDList))

                bigColonyIdx = popByArea.index(max(popByArea))
                powerColonyIdx = powerByArea.index(max(powerByArea))

                if( self.LOG_DEBUG ) : print ' Revolt: Largest colony is %d, with %d cities and %d population'%(areaIDList[bigColonyIdx],len(citiesInArea[bigColonyIdx]),popByArea[bigColonyIdx])
                avgCivPop = game.getTotalPopulation()/(1.0 * (game.countCivPlayersAlive() + game.countCivPlayersEverAlive())/2)
                if( self.LOG_DEBUG ) : print ' Revolt: Colony pop: %d, Adj world pop: %d (from %d)'%(popByArea[bigColonyIdx],avgCivPop,game.getTotalPopulation())
                powerPercent = int(round(powerByArea[powerColonyIdx]/(1.0*capital.area().getPower(player.getID()))*100))
                if( self.LOG_DEBUG ) : print ' Revolt: Most powerful colony is %d, with %d power (%d percent of homeland) and %d cities'%(areaIDList[powerColonyIdx],powerByArea[powerColonyIdx],powerPercent,len(citiesInArea[powerColonyIdx]))

                # Perhaps join nearby colonies?  So groups of islands secede together

        if( self.cultureSplit ) :
            if( self.LOG_DEBUG ) : print ' Revolt:  Attempting cultural split'
            idxForMostOwned = -1
            idxForMostCult = -1
            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                if( len(citiesOtherOwners[i]) > 0 and not i == player.getID() ) :
                    formerOwner = PyPlayer(i)
                    if( self.LOG_DEBUG ) : print ' Revolt:  Civ has %d cities formerly owned by the %s'%(len(citiesOtherOwners[i]),formerOwner.getCivilizationName())
                    if( idxForMostOwned < 0 ) :
                        idxForMostOwned = i
                    elif( len(citiesOtherOwners[i]) > len(citiesOtherOwners[idxForMostOwned]) ) :
                        idxForMostOwned = i
                if( len(citiesWithCulture[i]) > 0 and not i == player.getID() ) :
                    cultPlayer = PyPlayer(i)
                    if( self.LOG_DEBUG ) : print ' Revolt:  Civ has %d cities with culture or influenced by the %s'%(len(citiesWithCulture[i]),cultPlayer.getCivilizationName())
                    if( idxForMostCult < 0 ) :
                        idxForMostCult = i
                    elif( len(citiesWithCulture[i]) > len(citiesWithCulture[idxForMostCult]) ) :
                        idxForMostCult = i

            if( idxForMostCult > 0 and idxForMostOwned > 0 ) :
                if( self.LOG_DEBUG ) : print ' Revolt:  %s owned the most, %s have most culture influence'%(PyPlayer(idxForMostOwned).getCivilizationName(),PyPlayer(idxForMostCult).getCivilizationName())

        if( self.unhappySplit ) :
            if( self.LOG_DEBUG ) : print ' Revolt:  Attempting unhappy split'
            if( self.LOG_DEBUG ) : print ' Revolt:  Civ has %d unhappy cities'%(len(unhappyCities))

            if( len(unhappyCities) > 1 ) :

                closeCities = list()
                
                for i,baseCity in enumerate(unhappyCities) :
                    closeCities.append(list())
                    ix = baseCity.getX()
                    iy = baseCity.getY()
                    for j,city in enumerate(unhappyCities) :
                        if( not i == j ) :
                            jx = baseCity.getX()
                            jy = baseCity.getY()
                            if( plotDistance(ix,iy,jx,jy) < self.closeRadius ) :
                                closeCities[i].append(city)

                mostUnhappyNeighbors = 0
                idxForMost = -1
                for i,closeCityList in enumerate(closeCities) :
                    if( len(closeCityList) > mostUnhappyNeighbors ) :
                        idxForMost = i
                        mostUnhappyNeighbors = len(closeCityList)
                if( self.LOG_DEBUG ) : print ' Revolt:  Unhappy city %s has %d close unhappy cities'%(unhappyCities[idxForMost].getName(),mostUnhappyNeighbors)
                    

        if( self.empireSplit ) :
            if( self.LOG_DEBUG ) : print ' Revolt:  Attempting empire split'
            if( self.LOG_DEBUG ) : print ' Revolt:  Civ has %d cities in its homeland'%(len(homelandCities)+1) # Capital is not counted

            if( len(homelandCities) < 4 ) :
                if( self.LOG_DEBUG ) : print ' Revolt:  Not enough homeland cities for empire split'

            else :
                # Try split in x
                gameMap = gc.getMap()
                left = list()
                right = list()
                capX = capital.getX()
                
                if( gameMap.isWrapX() ) :
                    midX = gameMap.getGridWidth()/2
                    if( capX < midX ) : # in left half

                        mirrorX = capX + midX
                        for city in homelandCities :
                            if( city.getX() > mirrorX ) :
                                left.append(city)
                            elif( city.getX() > capX ) :
                                right.append(city)
                            else :
                                left.append(city)

                    else :
                        mirrorX = capX - midX
                        for city in homelandCities :
                            if( city.getX() < mirrorX ) :
                                right.append(city)
                            elif( city.getX() < capX ) :
                                left.append(city)
                            else :
                                right.append(city)
                else :
                    for city in homelandCities :
                        if( city.getX() > capX ) :
                            right.append(city)
                        else :
                            left.append(city)

                # Try split in y
                above = list()
                below = list()
                capY = capital.getY()
                
                if( gameMap.isWrapY() ) :
                    midY = gameMap.getGridHeight()/2
                    if( capY < midY ) : # in lower half

                        mirrorY = capY + midY
                        for city in homelandCities :
                            if( city.getY() > mirrorY ) :
                                below.append(city)
                            elif( city.getY() > capY ) :
                                above.append(city)
                            else :
                                below.append(city)

                    else :
                        mirrorY = capY - midY
                        for city in homelandCities :
                            if( city.getY() < mirrorY ) :
                                above.append(city)
                            elif( city.getY() < capY ) :
                                below.append(city)
                            else :
                                above.append(city)
                else :
                    for city in homelandCities :
                        if( city.getY() > capY ) :
                            above.append(city)
                        else :
                            below.append(city)

                # Compare which does a better job
                leftPop = 0
                for city in left :
                    leftPop += city.getPopulation()

                rightPop = 0
                for city in right :
                    rightPop += city.getPopulation()

                abovePop = 0
                for city in above :
                    abovePop += city.getPopulation()

                belowPop = 0
                for city in below :
                    belowPop += city.getPopulation()

                    
                Horiz = False
                Vert = False
                if( abs(leftPop - rightPop) <= abs(abovePop-belowPop) ) :
                    if( self.LOG_DEBUG ) : print ' Revolt:  Trying horizontal split'
                    Horiz = True
                    if( leftPop > rightPop ) :
                        big = left
                        small = right
                    else :
                        big = right
                        small = left

                else :
                    if( self.LOG_DEBUG ) : print ' Revolt:  Trying vertical split'
                    Vert = True
                    if( abovePop > belowPop ) :
                        big = above
                        small = below
                    else :
                        big = below
                        small = above

                if( self.LOG_DEBUG ) : print ' Revolt:  Num cities in large group: %d, in small: %d'%(len(big),len(small))

                smallPop = 0
                for city in small :
                    smallPop += city.getPopulation()

                if( smallPop > int(1.3*capital.getHighestPopulation()) ) :
                    CyInterface().addImmediateMessage("Splitting the %s!!!"%(player.getCivilizationDescription(0)),"")
                                        
                    if( self.LOG_DEBUG ) : print ' Revolt:  Proceeding with revolution!  Small pop is: %d'%(smallPop)

                    maxPop = 0
                    newCapital = None
                    for city in small :
                        if( city.getPopulation() > maxPop ) :
                            newCapital = city
                            maxPop = city.getPopulation()

                    if( self.LOG_DEBUG ) : print ' Revolt:  New capital chosen as  %s'%(newCapital.getName())

                    small.remove( newCapital )

                    newPlayer = self.spawnRevoltPlayer( newCapital.GetCy() )
                    if( newPlayer == None ) :
                        if( self.LOG_DEBUG ) : print ' Revolt:  Spawn player returned None, aborting revolution'
                        return False

                    popup = PyPopup.PyPopup(self.revolutionPopup,contextType = EventContextTypes.EVENTCONTEXT_ALL)
                    popup.setHeaderString( 'The %s is Collapsing!!!'%(player.getCivilizationDescription(0)) )
                    bodStr = 'With the fall of their capital, %s, the %s is splitting in two.'%(capital.getName(),player.getCivilizationDescription(0))
                    bodStr += '  The %s have been born, making their capital in %s!'%(newPlayer.getCivilizationDescription(0),newCapital.getName())
                    popup.setBodyString( bodStr )

                    popup.launch()

                    newPlayerID = newPlayer.getID()
                    newTeam = gc.getTeam( newPlayer.getTeam() )

                    newTeam.signOpenBorders( player.getTeam() )

                    newPlayer.acquireCity( newCapital.GetCy(), False, True )        

                    capital = newPlayer.getCapitalCity()
                    capital.setCulture( newPlayerID, capital.countTotalCulture(), True )

                    for city in small :
                        if( self.LOG_DEBUG ) : print ' Revolt:  Acquiring city %s'%(city.getName())
                        newPlayer.acquireCity( city.GetCy(), False, True )

                    newPlayerPy = PyPlayer( newPlayerID )

                    for city in newPlayerPy.getCityList() :
                        cityCy = city.GetCy()
                        if( not cityCy.isCapital() ) :
                            cityCy.setCulture( newPlayerID, cityCy.countTotalCulture()/2, True )

                    for unit in playerPy.getUnitList() :
                        # Transfer units to new player
                        if( unit.plot().getOwner() == newPlayerID ) :
                            if( unit.canGift(True) ) :
                                unit.gift()
                            
                else :
                    if( self.LOG_DEBUG ) : print ' Revolt:  Small pop is: %d, too small.  Aborting revolution.'%(smallPop)

                
        return False

    def spawnRevoltPlayer( self, pCity, giveTechs = True ) :
            
            owner = gc.getPlayer( pCity.getOwner() )

            if( owner.getID() == gc.getBARBARIAN_PLAYER() ) :
                return None
            
            cultOwner = None
            if( pCity.countTotalCulture() > 0 ) :
                cultOwner = gc.getPlayer( pCity.findHighestCulture() )
                if( cultOwner.getID() == gc.getBARBARIAN_PLAYER() ) :
                    cultOwner = None

            if( cultOwner == None ) :
                cultOwner = owner

            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - Spawning a revolt civ for the %s"%(cultOwner.getCivilizationDescription(0)))

            newPlayerIdx = -1
            # Overwrite a civ that never managed to found a city (probably a previous rebel)
            for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                if( (not gc.getPlayer(i).isAlive()) and (not gc.getPlayer(i).isFoundedFirstCity()) and (gc.getPlayer(i).getNumUnits() == 0) ) :
                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - Dead civ %d has founded city? %d"%(i,gc.getPlayer(i).isFoundedFirstCity()))
                    newPlayerIdx = i
                    break

            if( newPlayerIdx < 0 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - No 'fake' civs, looking for empty slots")
                # Put in empty slot
                for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                    if( (not gc.getPlayer(i).isEverAlive()) and (gc.getPlayer(i).getNumUnits() == 0) ) :
                        newPlayerIdx = i
                        break

            if( newPlayerIdx < 0 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - No available slots (%d)"%(newPlayerIdx))
                return None

            # Don't reincarnate as either of these
            iMinor = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_MINOR')
            iBarbarian = CvUtil.findInfoTypeNum(gc.getCivilizationInfo,gc.getNumCivilizationInfos(),'CIVILIZATION_BARBARIAN')
            # Civs not currently in the game
            availableCivs = list()
            # Civs with similar style to cultOwner, if they exist
            similarStyleCivs = list()
            for civType in range(0,gc.getNumCivilizationInfos()) :
                if( not civType == iBarbarian ) :
                    if( not civType == iMinor ) :
                        taken = False
                        for i in range(0,gc.getMAX_CIV_PLAYERS()) :
                            if( civType == gc.getPlayer(i).getCivilizationType() ) :
                                if( gc.getPlayer(i).isAlive() or gc.getPlayer(i).isFoundedFirstCity() ) :
                                    taken = True
                                    break
                        if( not taken ) :
                            availableCivs.append(civType)
                            if( not cultOwner == None ) :
                                if( gc.getCivilizationInfo( cultOwner.getCivilizationType() ).getArtStyleType() == gc.getCivilizationInfo(civType).getArtStyleType() ) :
                                    if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - Potential similar style civ: %s"%(gc.getCivilizationInfo(civType).getShortDescription(0)))
                                    similarStyleCivs.append(civType)

            if( len(availableCivs) < 1 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - No available civs")
                return None

            if( len(similarStyleCivs) > 0 ) :
                # Create a civ using this style
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - Creating a similar style civ to %s"%(cultOwner.getCivilizationDescription(0)))
                availableCivs = similarStyleCivs

            newCivIdx = availableCivs[game.getSorenRandNum(len(availableCivs),'Revolt: pick unused civ type')]

            leaderList = list()
            for leaderType in range(0,gc.getNumLeaderHeadInfos()) :
                if( gc.getCivilizationInfo(newCivIdx).isLeaders(leaderType) ) :
                    leaderList.append(leaderType)

            if( len(leaderList) < 1 ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - Unexpected lack of possible leaders")
                return None
            
            newLeaderIdx = leaderList[game.getSorenRandNum(len(leaderList),'Revolt: pick leader')]

            success = game.createNewPlayer( newPlayerIdx, newCivIdx, newLeaderIdx, -1, False, 1, 1 )

            if( not success ) :
                if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - New civ creation failed")
                return None

            pRevolt = gc.getPlayer(newPlayerIdx)

            if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - Revolt civ %s spawned in slot %d"%(pRevolt.getCivilizationDescription(0),newPlayerIdx))

            if( giveTechs ) :
                # Give all techs known by city owner, except a few of the most expensive
                knownTechs = list()
                mostExpensive = list()
                minMostExpensive = 0
                numMostExpensive = 2 + game.getSorenRandNum(4,'Revolt')

                ownerTeam = gc.getTeam( owner.getID() )
                
                for techID in range(0,gc.getNumTechInfos()) :
                    if( ownerTeam.isHasTech( techID ) ) :
                        knownTechs.append( techID )
                        if( gc.getTechInfo( techID ).getResearchCost() > minMostExpensive ) :
                            if( len(mostExpensive) < numMostExpensive ) :
                                mostExpensive.append( techID )
                            else :
                                for j in range(0,len(mostExpensive)) :
                                    if( gc.getTechInfo( mostExpensive[j] ).getResearchCost() == minMostExpensive ) :
                                        mostExpensive.remove( mostExpensive[j] )
                                        break
                                mostExpensive.append( techID )
                                minMostExpensive = gc.getTechInfo( mostExpensive[0] ).getResearchCost()
                                for j in range(0,len(mostExpensive)) :
                                    if( gc.getTechInfo( mostExpensive[j] ).getResearchCost() < minMostExpensive ) :
                                        minMostExpensive = gc.getTechInfo( mostExpensive[j] ).getResearchCost()

                revoltTeam = gc.getTeam( pRevolt.getID() )
                for techID in knownTechs :
                    if( not techID in mostExpensive ) :
                        revoltTeam.setHasTech(techID,True,pRevolt.getID(),False,False)
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - Giving rebel %s"%(PyInfo.TechnologyInfo(techID).getDescription()))
                    else :
                        if( self.LOG_DEBUG ) : CvUtil.pyPrint("Revolt - NOT giving rebel %s"%(PyInfo.TechnologyInfo(techID).getDescription()))

            return pRevolt
